# -*- coding: UTF-8 -*-
__author__ = "Jaime Rivera <jaime.rvq@gmail.com>"
__copyright__ = "Copyright 2022, Jaime Rivera"
__credits__ = []
__license__ = "MIT License"


import logging
from functools import partial
import math
import pprint

from PySide2 import QtWidgets
from PySide2 import QtCore
from PySide2 import QtGui
from PySide2 import QtSvg

from all_nodes import constants
from all_nodes import utils
from all_nodes.graphic.widgets.global_signaler import GlobalSignaler


GS = GlobalSignaler()

LOGGER = utils.get_logger(__name__)


# -------------------------------- NODE CLASS -------------------------------- #
class GeneralGraphicNode(QtWidgets.QGraphicsPathItem):
    def __init__(self, logic_node, color_name):
        # LOGIC NODE
        self.logic_node = logic_node

        # SHAPE
        self.extra_header = 0
        self.extra_bottom = 0
        self.node_width = 450

        # INIT
        QtWidgets.QGraphicsPathItem.__init__(self)
        self.setData(0, constants.GRAPHIC_NODE)
        self.setAcceptHoverEvents(True)
        self.setFlags(
            QtWidgets.QGraphicsPathItem.ItemIsMovable
            | QtWidgets.QGraphicsPathItem.ItemIsSelectable
            | QtWidgets.QGraphicsPathItem.ItemSendsScenePositionChanges
        )

        # COLOR
        self.base_color = QtGui.QColor(color_name)
        self.bright_color_name = utils.get_bright_color(color_name)

        # SUB-ITEMS
        if "svg" in self.logic_node.ICON_PATH:
            self.class_icon_renderer = QtSvg.QSvgRenderer(self.logic_node.ICON_PATH)
            self.class_icon = QtSvg.QGraphicsSvgItem(parentItem=self)
            self.class_icon.setSharedRenderer(self.class_icon_renderer)
            self.class_icon.setScale(
                constants.STRIPE_HEIGHT / self.class_icon.boundingRect().height()
            )
        else:
            class_pixmap = QtGui.QPixmap(self.logic_node.ICON_PATH)
            class_pixmap = class_pixmap.scaledToWidth(
                constants.STRIPE_HEIGHT,
                QtCore.Qt.TransformationMode.SmoothTransformation,
            )
            self.class_icon = QtWidgets.QGraphicsPixmapItem(class_pixmap, parent=self)

        self.class_text = QtWidgets.QGraphicsTextItem(parent=self)

        self.proxy_help_btn = QtWidgets.QGraphicsProxyWidget(parent=self)

        self.stripes = []

        self.selection_marquee = QtWidgets.QGraphicsPathItem(parent=self)
        self.selection_marquee.hide()

        self.deactivated_cross = QtWidgets.QGraphicsPathItem(parent=self)
        self.deactivated_cross.hide()

        self.error_marquee = QtWidgets.QGraphicsPathItem(parent=self)
        self.error_marquee.setFlag(QtWidgets.QGraphicsItem.ItemStacksBehindParent)
        self.error_marquee.hide()

        self.additional_info_text = QtWidgets.QGraphicsTextItem(parent=self)

        self.glow = QtWidgets.QGraphicsPathItem(parent=self)
        self.glow.hide()

        self.badge_renderer = QtSvg.QSvgRenderer("icons:badges.svg")
        self.badge_icon = QtSvg.QGraphicsSvgItem(parentItem=self)
        self.badge_icon.setSharedRenderer(self.badge_renderer)
        self.badge_icon.setElementId("neutral")
        self.badge_icon.hide()

        self.extras_renderer = QtSvg.QSvgRenderer("icons:ctx.svg")

        # ATTRIBUTES
        self.graphic_attributes = []

        # DIRECT INPUT WIDGETS AND PREVIEWS
        self.input_widgets = []
        self.proxy_input_widgets = []

        self.preview_widgets = []
        self.proxy_preview_widgets = []

        self.widget_registry = []

        # SETUP
        self.setup_node()
        self.setup_graphic_attributes()
        self.setup_input_widgets()
        self.setup_preview_widgets()
        self.setup_extras()

        self.make_shape()
        self.place_graphic_attributes()
        self.place_widgets()

        self.update_attributes_from_widgets()

        # Connect
        self.logic_node.signaler.is_executing.connect(self.show_executing)
        self.logic_node.signaler.status_changed.connect(self.show_result)
        self.logic_node.signaler.finished.connect(self.show_result)

    # PROPERTIES ----------------------
    @property
    def node_height(self):
        total_h = (
            constants.HEADER_HEIGHT
            + self.extra_header
            + (self.logic_node.get_max_in_or_out_count()) * constants.STRIPE_HEIGHT
            + self.extra_bottom
        )
        total_h += constants.STRIPE_HEIGHT

        return total_h

    @property
    def has_gui_inputs(self):
        if self.logic_node.get_gui_internals_inputs():
            return True

        return False

    @property
    def has_gui_previews(self):
        if self.logic_node.get_gui_internals_previews():
            return True

        return False

    # GRAPHICS SETUP ----------------------
    def guess_width_to_use(self):
        """
        Calculate width to draw the node with, based on the size of its graphical attributes.
        """
        # Biggest attr names
        longest_in_name = max(
            [
                attribute.attr_text.boundingRect().width()
                for attribute in self.graphic_attributes
                if attribute.connector_type == constants.INPUT
            ]
        )
        longest_out_name = max(
            [
                attribute.attr_text.boundingRect().width()
                for attribute in self.graphic_attributes
                if attribute.connector_type == constants.OUTPUT
            ]
        )

        # Calculate max
        self.node_width = max(
            2.5 * constants.CHAMFER_RADIUS
            + self.class_text.boundingRect().width()
            + self.proxy_help_btn.boundingRect().width()
            + 20,
            longest_in_name + longest_out_name + 2 * constants.CHAMFER_RADIUS + 20,
        )

        if self.has_gui_inputs:
            self.node_width = max(
                self.node_width,
                max(
                    w.width() + 2 * constants.CHAMFER_RADIUS for w in self.input_widgets
                ),
            )

        if self.has_gui_previews:
            self.node_width = max(
                self.node_width,
                max(
                    w.width() + 2 * constants.CHAMFER_RADIUS
                    for w in self.preview_widgets
                ),
            )

    def setup_node(self):
        """
        Setup some of the subcomponents of the node.
        """
        # NODE ICON
        self.class_icon.moveBy(
            constants.CHAMFER_RADIUS / 2.0, constants.CHAMFER_RADIUS / 2.0
        )

        # NODE NAME
        self.class_text.setHtml(
            '<p align="left"><font color=white>{0}'.format(self.logic_node.node_name)
        )
        node_class_font = QtGui.QFont(
            constants.NODE_FONT, int(0.4 * constants.HEADER_HEIGHT)
        )
        self.class_text.setFont(node_class_font)
        self.class_text.moveBy(2.5 * constants.CHAMFER_RADIUS, 0)

        # HELP
        help_btn = QtWidgets.QPushButton(parent=None)
        help_btn.setStyleSheet(
            "QPushButton:hover{color:yellow; background-color:rgba(255,255,0,50)}"
            "QPushButton{color:white; border-radius:3px; border:1px solid white; background:transparent;}"
        )
        help_btn.setFixedSize(
            int(0.4 * constants.HEADER_HEIGHT), int(0.4 * constants.HEADER_HEIGHT)
        )
        help_btn.setText("?")
        help_btn.setFont(
            QtGui.QFont(constants.NODE_FONT, int(0.45 * help_btn.height()))
        )
        help_btn.clicked.connect(self.show_help)
        self.proxy_help_btn = QtWidgets.QGraphicsProxyWidget(parent=self)
        self.proxy_help_btn.setWidget(help_btn)

        # ADDITIONAL INFO
        node_id_font = QtGui.QFont(
            constants.NODE_FONT, int(0.5 * constants.STRIPE_HEIGHT)
        )
        self.additional_info_text.setFont(node_id_font)
        self.additional_info_text.hide()

    def setup_graphic_attributes(self):
        """
        Create graphic attributes of this node
        """
        i = 0
        for attr in self.logic_node.get_input_attrs():
            attr = GeneralGraphicAttribute(attr, self, i)
            attr.setZValue(100)  # TODO maybe put all the setZValue together?
            self.graphic_attributes.append(attr)
            i += 1

        i = 0
        for attr in self.logic_node.get_output_attrs():
            attr = GeneralGraphicAttribute(attr, self, i)
            attr.setZValue(100)
            self.graphic_attributes.append(attr)
            i += 1

    def add_single_graphic_attribute(
        self,
        attribute_name,
        connector_type,
        data_type,
        gui_type=None,
        is_optional=False,
        value=None,
    ):
        # Check if it can be added
        new_logic_attr = self.logic_node.add_attribute(
            attribute_name, connector_type, data_type, gui_type, is_optional, value
        )
        if not new_logic_attr:
            GS.signals.main_screen_feedback.emit(
                "Cannot create attribute, name '{}' already exists in the node".format(
                    attribute_name
                ),
                logging.ERROR,
            )
            return

        # Create graphic attribute
        if connector_type in [constants.INPUT, constants.OUTPUT]:
            max_row = len(
                [
                    attribute
                    for attribute in self.graphic_attributes
                    if attribute.connector_type == connector_type
                ]
            )
            attr = GeneralGraphicAttribute(new_logic_attr, self, max_row)
            attr.setZValue(100)  # TODO maybe put all the setZValue together?
            self.graphic_attributes.append(attr)

        elif connector_type == constants.INTERNAL:
            self.setup_input_widgets()
            self.setup_preview_widgets()

        self.make_shape()
        self.place_graphic_attributes()
        self.place_widgets()

        # Feedback
        GS.signals.main_screen_feedback.emit(
            "Added new attribute '{}' to the node".format(attribute_name), logging.INFO
        )

    def make_shape(self):
        """
        Setup the shape and size of the node and its subcomponents.
        """
        # ESTABLISH WIDTH OF THE NODE
        self.guess_width_to_use()

        # BASIC SHAPE
        self.setZValue(20)
        self.setPen(
            QtGui.QPen(
                QtGui.QColor(self.bright_color_name), constants.NODE_CONTOUR_THICKNESS
            )
        )
        new_path = QtGui.QPainterPath()
        new_path.addRoundedRect(
            QtCore.QRect(0, 0, self.node_width, self.node_height),
            constants.CHAMFER_RADIUS,
            constants.CHAMFER_RADIUS,
        )
        self.setPath(new_path)

        # FILLING
        grad = QtGui.QLinearGradient(0, 0, 0, self.node_height)
        grad.setColorAt(0.5, self.base_color)
        lighter_color = QtGui.QColor(
            self.base_color.red() + 50,
            self.base_color.green() + 50,
            self.base_color.blue() + 50,
        )
        grad.setColorAt(1.0, lighter_color)
        self.setBrush(grad)

        # STRIPES
        for stripe in self.stripes:
            stripe.setParentItem(None)
            del stripe
        stripe_path = QtGui.QPainterPath()
        stripe_path.addRect(
            QtCore.QRect(0, 0, self.node_width, constants.STRIPE_HEIGHT)
        )
        for row_count in range(self.logic_node.get_max_in_or_out_count()):
            stripe = QtWidgets.QGraphicsPathItem(parent=self)
            stripe.setPath(stripe_path)
            stripe.setPen(QtGui.QPen(QtCore.Qt.NoPen))
            stripe.moveBy(
                0,
                constants.HEADER_HEIGHT
                + self.extra_header
                + row_count * constants.STRIPE_HEIGHT,
            )
            if row_count % 2 == 0:
                stripe.setBrush(QtGui.QColor(255, 255, 255, 25))
            stripe.setZValue(10)
            self.stripes.append(stripe)

        # BUTTON MOVEMENT
        self.proxy_help_btn.setPos(
            self.node_width
            - self.proxy_help_btn.widget().width()
            - constants.CHAMFER_RADIUS / 2,
            constants.CHAMFER_RADIUS / 2,
        )

        # GLOW
        glow_pen = QtGui.QPen(
            QtGui.QColor(self.bright_color_name), constants.NODE_SELECTED_GLOW_THICKNESS
        )
        self.glow.setPen(glow_pen)
        self.glow.setPath(new_path)
        blur = QtWidgets.QGraphicsBlurEffect()
        blur.setBlurRadius(glow_pen.width() * 2)
        self.glow.setGraphicsEffect(blur)
        self.glow.setZValue(40)

        # SELECTION MARQUEE
        self.selection_marquee.setPath(new_path)
        self.selection_marquee.setPen(constants.NODE_SELECTED_PEN)
        self.selection_marquee.setZValue(50)

        # DEACTIVATED CROSS
        x_path = QtGui.QPainterPath()
        x_path.lineTo(
            QtCore.QPoint(
                self.node_width,
                self.node_height,
            ),
        )
        x_path.moveTo(
            QtCore.QPoint(0, self.node_height),
        )
        x_path.lineTo(
            QtCore.QPoint(
                self.node_width,
                0,
            ),
        )
        self.deactivated_cross.setPath(x_path)
        self.deactivated_cross.setPen(constants.NODE_DEACTIVATED_PEN)
        self.deactivated_cross.setZValue(300)

        # ERROR MARQUEE
        error_path = QtGui.QPainterPath()
        error_path.addRoundedRect(
            QtCore.QRect(
                0,
                0,
                self.node_width + constants.PLUG_RADIUS * 6,
                self.node_height + constants.PLUG_RADIUS * 6,
            ),
            constants.CHAMFER_RADIUS * 2,
            constants.CHAMFER_RADIUS * 2,
        )
        self.error_marquee.setPath(error_path)
        self.error_marquee.moveBy(
            -constants.PLUG_RADIUS * 3, -constants.PLUG_RADIUS * 3
        )

        # BADGES
        self.badge_icon.setZValue(80)
        self.badge_icon.setPos(
            self.node_width
            - self.badge_icon.boundingRect().width()
            - constants.CHAMFER_RADIUS,
            self.node_height - constants.STRIPE_HEIGHT / 2,
        )

        # EXTRA TEXT
        self.additional_info_text.setTextWidth(self.node_width * 1.4)
        self.additional_info_text.moveBy(
            -constants.CHAMFER_RADIUS,
            self.node_height + constants.PLUG_RADIUS * 3,
        )

    def place_graphic_attributes(self):
        """
        Place the graphic attributes in their correct position.
        """
        for attr in self.graphic_attributes:
            attr.setup_graphics()

    def place_widgets(self):
        """
        Place the input and preview widgets in their correct position.
        """
        if self.has_gui_inputs:
            for i in range(len(self.input_widgets)):
                self.input_widgets[i].setFixedSize(
                    max(
                        self.node_width - 2 * constants.CHAMFER_RADIUS,
                        self.input_widgets[i].width(),
                    ),
                    self.input_widgets[i].height(),
                )

                self.proxy_input_widgets.append(
                    QtWidgets.QGraphicsProxyWidget(parent=self)
                )
                self.proxy_input_widgets[i].setWidget(self.input_widgets[i])
                self.proxy_input_widgets[i].setPos(
                    constants.CHAMFER_RADIUS, constants.HEADER_HEIGHT
                )
                self.proxy_input_widgets[i].setZValue(
                    200
                )  # TODO maybe put all the setZValue together?
                if i:
                    in_widget = self.proxy_input_widgets[i - 1].widget()
                    self.proxy_input_widgets[i].setPos(
                        constants.CHAMFER_RADIUS,
                        in_widget.pos().y() + in_widget.height(),
                    )

        if self.has_gui_previews:
            for i in range(len(self.preview_widgets)):
                self.preview_widgets[i].setFixedSize(
                    max(
                        self.node_width - 2 * constants.CHAMFER_RADIUS,
                        self.preview_widgets[i].width(),
                    ),
                    self.preview_widgets[i].height(),
                )

                self.proxy_preview_widgets.append(
                    QtWidgets.QGraphicsProxyWidget(parent=self)
                )
                self.proxy_preview_widgets[i].setWidget(self.preview_widgets[i])
                self.proxy_preview_widgets[i].setPos(
                    constants.CHAMFER_RADIUS,
                    constants.HEADER_HEIGHT
                    + self.extra_header
                    + (self.logic_node.get_max_in_or_out_count())
                    * constants.STRIPE_HEIGHT
                    + constants.CHAMFER_RADIUS,
                )
                if i:
                    prev_widget = self.proxy_preview_widgets[i - 1].widget()
                    self.proxy_preview_widgets[i].setPos(
                        constants.CHAMFER_RADIUS,
                        prev_widget.pos().y() + prev_widget.height(),
                    )

    def setup_input_widgets(self):
        """
        For special nodes that take input, setup widgets to receive the input
        """
        if not self.has_gui_inputs:
            return

        # Set type of widget depending on the type of input neede
        gui_internals_inputs = self.logic_node.get_gui_internals_inputs()

        for attr_name in gui_internals_inputs:
            # Check if a widget was already created
            if attr_name in self.widget_registry:
                continue
            else:
                self.widget_registry.append(attr_name)

            # Setup a widget for this attr
            gui_input_type = gui_internals_inputs.get(attr_name).get("gui_type")

            if gui_input_type == constants.InputsGUI.STR_INPUT:
                new_input_widget = QtWidgets.QLineEdit(parent=None)
                new_input_widget.setObjectName(attr_name)
                new_input_widget.setToolTip(f"Internal attribute: {attr_name}")
                new_input_widget.setFixedSize(
                    150,
                    int(constants.HEADER_HEIGHT),
                )
                new_input_widget.setStyleSheet(
                    "background:transparent; color:white; border:1px solid white;"
                )
                new_input_widget.setPlaceholderText("str here")
                if self.logic_node.get_attribute_value(attr_name):
                    new_input_widget.setText(
                        self.logic_node.get_attribute_value(attr_name)
                    )
                new_input_widget.textChanged.connect(
                    partial(
                        self.update_attributes_from_widgets,
                        attr_name,
                    )
                )
                self.input_widgets.append(new_input_widget)

            elif gui_input_type == constants.InputsGUI.MULTILINE_STR_INPUT:
                new_input_widget = QtWidgets.QPlainTextEdit(parent=None)
                new_input_widget.setObjectName(attr_name)
                new_input_widget.setToolTip(f"Internal attribute: {attr_name}")
                new_input_widget.setFixedSize(
                    300,
                    int(4 * constants.HEADER_HEIGHT),
                )
                new_input_widget.setStyleSheet(
                    "background:transparent; color:white; border:1px solid white;"
                )
                new_input_widget.setPlaceholderText("Text here")
                if self.logic_node.get_attribute_value(attr_name):
                    new_input_widget.setPlainText(
                        self.logic_node.get_attribute_value(attr_name)
                    )
                new_input_widget.textChanged.connect(
                    partial(
                        self.update_attributes_from_widgets,
                        attr_name,
                    )
                )
                self.input_widgets.append(new_input_widget)

            elif gui_input_type == constants.InputsGUI.DICT_INPUT:
                new_input_widget = QtWidgets.QPlainTextEdit(parent=None)
                new_input_widget.setObjectName(attr_name)
                new_input_widget.setToolTip(f"Internal attribute: {attr_name}")
                new_input_widget.setFixedSize(
                    300,
                    int(4 * constants.HEADER_HEIGHT),
                )
                new_input_widget.setStyleSheet(
                    "background:transparent; color:white; border:1px solid white;"
                )
                new_input_widget.setPlaceholderText("Text here")
                if self.logic_node.get_attribute_value(attr_name):
                    new_input_widget.setText(
                        self.logic_node.get_attribute_value(attr_name)
                    )
                new_input_widget.textChanged.connect(
                    partial(
                        self.update_attributes_from_widgets,
                        attr_name,
                    )
                )
                self.input_widgets.append(new_input_widget)

            elif gui_input_type == constants.InputsGUI.LIST_INPUT:
                new_input_widget = QtWidgets.QLineEdit(parent=None)
                new_input_widget.setObjectName(attr_name)
                new_input_widget.setToolTip(f"Internal attribute: {attr_name}")
                new_input_widget.setFixedSize(
                    150,
                    int(constants.HEADER_HEIGHT),
                )
                new_input_widget.setStyleSheet(
                    "background:transparent; color:white; border:1px solid white;"
                )
                new_input_widget.setPlaceholderText("list here")
                if self.logic_node.get_attribute_value(attr_name):
                    new_input_widget.setText(
                        str(self.logic_node.get_attribute_value(attr_name))
                    )
                new_input_widget.textChanged.connect(
                    partial(
                        self.update_attributes_from_widgets,
                        attr_name,
                    )
                )
                self.input_widgets.append(new_input_widget)

            elif gui_input_type == constants.InputsGUI.BOOL_INPUT:
                new_input_widget = QtWidgets.QCheckBox("False", parent=None)
                new_input_widget.setObjectName(attr_name)
                new_input_widget.setToolTip(f"Internal attribute: {attr_name}")
                new_input_widget.setFixedSize(150, int(constants.HEADER_HEIGHT))
                new_input_widget.setStyleSheet(
                    "QCheckBox::indicator{border : 1px solid white;}"
                    "QCheckBox::indicator:checked{ background:rgba(255,255,200,150); }"
                    "QCheckBox{ background:transparent; color:white}"
                )
                if self.logic_node.get_attribute_value(attr_name):
                    new_input_widget.setChecked(
                        self.logic_node.get_attribute_value(attr_name)
                    )
                new_input_widget.stateChanged.connect(
                    lambda: new_input_widget.setText(
                        ["False", "True"][new_input_widget.isChecked()]
                    )
                )
                new_input_widget.clicked.connect(
                    partial(self.update_attributes_from_widgets, attr_name)
                )
                self.input_widgets.append(new_input_widget)

            elif gui_input_type == constants.InputsGUI.INT_INPUT:
                new_input_widget = QtWidgets.QSpinBox(parent=None)
                new_input_widget.setObjectName(attr_name)
                new_input_widget.setToolTip(f"Internal attribute: {attr_name}")
                new_input_widget.setFixedSize(
                    150,
                    int(constants.HEADER_HEIGHT),
                )
                new_input_widget.setStyleSheet(
                    "background:transparent; color:white; border:1px solid white;"
                )
                new_input_widget.setMaximum(int(1e6))
                new_input_widget.setMinimum(int(-1e6))
                if self.logic_node.get_attribute_value(attr_name):
                    new_input_widget.setValue(
                        self.logic_node.get_attribute_value(attr_name)
                    )
                new_input_widget.valueChanged.connect(
                    partial(self.update_attributes_from_widgets, attr_name)
                )
                self.input_widgets.append(new_input_widget)

            elif gui_input_type == constants.InputsGUI.FLOAT_INPUT:
                new_input_widget = QtWidgets.QDoubleSpinBox(parent=None)
                new_input_widget.setObjectName(attr_name)
                new_input_widget.setToolTip(f"Internal attribute: {attr_name}")
                new_input_widget.setFixedSize(
                    150,
                    int(constants.HEADER_HEIGHT),
                )
                new_input_widget.setStyleSheet(
                    "background:transparent; color:white; border:1px solid white;"
                )
                new_input_widget.setMaximum(int(1e6))
                new_input_widget.setMinimum(int(-1e6))
                if self.logic_node.get_attribute_value(attr_name):
                    new_input_widget.setValue(
                        self.logic_node.get_attribute_value(attr_name)
                    )
                new_input_widget.valueChanged.connect(
                    partial(self.update_attributes_from_widgets, attr_name)
                )
                self.input_widgets.append(new_input_widget)

            elif gui_input_type == constants.InputsGUI.OPTION_INPUT:
                new_input_widget = QtWidgets.QComboBox(parent=None)
                new_input_widget.setObjectName(attr_name)
                new_input_widget.setToolTip(f"Internal attribute: {attr_name}")
                new_input_widget.setFixedSize(
                    150,
                    int(constants.HEADER_HEIGHT),
                )
                new_input_widget.adjustSize()
                new_input_widget.setFixedSize(
                    new_input_widget.width() + 20, new_input_widget.height()
                )
                new_input_widget.setStyleSheet(
                    "QComboBox { background:transparent; color:white; border:1px solid white; }"
                    "QWidget:item { color: black; background:white; }"
                )
                new_input_widget.addItems(
                    gui_internals_inputs.get(attr_name).get(
                        "options", constants.InputsGUI.OPTION_INPUT.value
                    )
                )
                if self.logic_node.get_attribute_value(attr_name):
                    new_input_widget.setCurrentText(
                        self.logic_node.get_attribute_value(attr_name)
                    )
                new_input_widget.currentIndexChanged.connect(
                    partial(self.update_attributes_from_widgets, attr_name)
                )
                self.input_widgets.append(new_input_widget)

            elif gui_input_type == constants.InputsGUI.TUPLE_INPUT:
                new_input_widget = QtWidgets.QLineEdit(parent=None)
                new_input_widget.setObjectName(attr_name)
                new_input_widget.setToolTip(f"Internal attribute: {attr_name}")
                new_input_widget.setFixedSize(
                    150,
                    int(constants.HEADER_HEIGHT),
                )
                new_input_widget.setStyleSheet(
                    "background:transparent; color:white; border:1px solid white;"
                )
                new_input_widget.setPlaceholderText("tuple here")
                if self.logic_node.get_attribute_value(attr_name):
                    new_input_widget.setText(
                        self.logic_node.get_attribute_value(attr_name)
                    )
                new_input_widget.textChanged.connect(
                    partial(
                        self.update_attributes_from_widgets,
                        attr_name,
                    )
                )
                self.input_widgets.append(new_input_widget)

        # Set font of the widgets
        for w in self.input_widgets:
            w.setFont(QtGui.QFont(constants.NODE_FONT, constants.HEADER_HEIGHT * 0.4))

        # Measure
        self.extra_header = sum([w.height() for w in self.input_widgets]) + 5

    def setup_preview_widgets(self):
        """
        For special nodes that can display previews, setup widgets
        """
        if not self.has_gui_previews:
            return

        # Separator label
        if not self.widget_registry:
            previews_label = QtWidgets.QLabel(parent=None)
            previews_label.setFixedSize(
                150,
                int(constants.STRIPE_HEIGHT),
            )
            previews_label.setStyleSheet(
                "background:transparent; color:white; border:none;"
            )
            previews_label.setText("- PREVIEWS -")
            self.preview_widgets.append(previews_label)

        # Set type of widget depending on the type of preview needed
        gui_internals_previews = self.logic_node.get_gui_internals_previews()

        for attr_name in gui_internals_previews:
            # Check if a widget was already created
            if attr_name in self.widget_registry:
                continue
            else:
                self.widget_registry.append(attr_name)

            # Create widget for attribute
            gui_preview_type = gui_internals_previews.get(attr_name).get("gui_type")

            if gui_preview_type == constants.PreviewsGUI.STR_PREVIEW:
                new_preview_widget = QtWidgets.QLineEdit(parent=None)
                new_preview_widget.setObjectName(attr_name)
                new_preview_widget.setToolTip(f"Internal attribute: {attr_name}")
                new_preview_widget.setReadOnly(True)
                new_preview_widget.setFixedSize(
                    150,
                    int(constants.STRIPE_HEIGHT),
                )
                new_preview_widget.setStyleSheet(
                    "background:transparent; color:white; border:1px dotted white;"
                )

                if self.logic_node.get_attribute_value(attr_name):
                    new_preview_widget.setText(
                        self.logic_node.get_attribute_value(attr_name)
                    )
                else:
                    new_preview_widget.setText(f"[{attr_name}]")

                self.preview_widgets.append(new_preview_widget)

            elif gui_preview_type == constants.PreviewsGUI.MULTILINE_STR_PREVIEW:
                new_preview_widget = QtWidgets.QPlainTextEdit(parent=None)
                new_preview_widget.setObjectName(attr_name)
                new_preview_widget.setToolTip(f"Internal attribute: {attr_name}")
                new_preview_widget.setReadOnly(True)
                new_preview_widget.setFixedSize(
                    300,
                    int(4 * constants.HEADER_HEIGHT),
                )
                new_preview_widget.setStyleSheet(
                    "background:transparent; color:white; border:1px dotted white;"
                )

                if self.logic_node.get_attribute_value(attr_name):
                    new_preview_widget.setPlaceholderText(
                        self.logic_node.get_attribute_value(attr_name)
                    )
                else:
                    new_preview_widget.setPlaceholderText(f"[{attr_name}]")

                self.preview_widgets.append(new_preview_widget)

            elif gui_preview_type == constants.PreviewsGUI.DICT_PREVIEW:
                new_preview_widget = QtWidgets.QPlainTextEdit(parent=None)
                new_preview_widget.setObjectName(attr_name)
                new_preview_widget.setToolTip(f"Internal attribute: {attr_name}")
                new_preview_widget.setReadOnly(True)
                new_preview_widget.setFixedSize(
                    300,
                    int(4 * constants.HEADER_HEIGHT),
                )
                new_preview_widget.setStyleSheet(
                    "background:transparent; color:white; border:1px dotted white;"
                )

                if self.logic_node.get_attribute_value(attr_name):
                    new_preview_widget.setPlaceholderText(
                        self.logic_node.get_attribute_value(attr_name)
                    )
                else:
                    new_preview_widget.setPlaceholderText(f"[{attr_name}]")

                self.preview_widgets.append(new_preview_widget)

            elif gui_preview_type == constants.PreviewsGUI.IMAGE_PREVIEW:
                new_preview_widget = QtWidgets.QLabel(parent=None)
                new_preview_widget.setObjectName(attr_name)
                new_preview_widget.setToolTip(f"Internal attribute: {attr_name}")
                new_preview_widget.setFixedSize(
                    constants.STRIPE_HEIGHT * 10,
                    constants.STRIPE_HEIGHT * 10,
                )
                new_preview_widget.setAlignment(
                    QtCore.Qt.AlignVCenter and QtCore.Qt.AlignCenter
                )
                new_preview_widget.setStyleSheet(
                    "background:transparent; color:white; border:none;"
                )
                if self.logic_node.get_attribute_value(attr_name):
                    pass  # TODO
                else:
                    new_preview_widget.setText(f"[{attr_name}]")

                self.preview_widgets.append(new_preview_widget)

        # Measure
        self.extra_bottom = sum([w.height() for w in self.preview_widgets])

    def setup_extras(self):
        """
        Setup the extra labels/icons on the node.

        Mainly oriented to context nodes or context-related nodes.
        """
        if self.logic_node.IS_CONTEXT:
            ctx_icon = QtSvg.QGraphicsSvgItem(parentItem=self)
            ctx_icon.setSharedRenderer(self.extras_renderer)
            ctx_icon.setPos(0, -25)
            ctx_icon.setElementId("context")

        elif "InputFromCtx" in self.logic_node.class_name:
            ctx_icon = QtSvg.QGraphicsSvgItem(parentItem=self)
            ctx_icon.setSharedRenderer(self.extras_renderer)
            ctx_icon.setPos(0, -25)
            ctx_icon.setElementId("in")

        elif "OutputToCtx" in self.logic_node.class_name:
            ctx_icon = QtSvg.QGraphicsSvgItem(parentItem=self)
            ctx_icon.setSharedRenderer(self.extras_renderer)
            ctx_icon.setPos(0, -25)
            ctx_icon.setElementId("out")

    # EXECUTION ----------------------
    def reset(self):
        """
        Reset the appearance of the node, so it shows no execution feedback (badges, marquees...).
        """
        LOGGER.info("Resetting graphic node {}".format(self.logic_node.node_name))
        self.badge_icon.hide()
        self.error_marquee.hide()
        self.additional_info_text.hide()
        self.update_attributes_from_widgets()
        self.clear_previews()

    def show_executing(self):
        """
        Display visual indications around the node to show it is running.
        """
        self.badge_icon.show()

        self.badge_icon.setElementId("executing")
        self.badge_icon.setToolTip('<p style="color: magenta">Executing...<br>')

        self.error_marquee.hide()

    def show_result(self):
        """
        Display visual indications around the node to show the result of its execution.
        """
        self.badge_icon.show()

        if self.logic_node.success == constants.NOT_RUN:
            self.badge_icon.hide()
            self.error_marquee.hide()

        elif self.logic_node.success == constants.IN_LOOP:
            self.badge_icon.setElementId("loop")
            self.badge_icon.setToolTip('<p style="color: cyan">In loop<br>')

            self.error_marquee.hide()

        elif self.logic_node.success == constants.SKIPPED:
            self.badge_icon.setElementId("neutral")
            self.badge_icon.setToolTip('<p style="color: gray">Skipped<br>')

            self.error_marquee.hide()

        elif self.logic_node.success == constants.SUCCESSFUL:
            self.badge_icon.setElementId("okay")
            self.badge_icon.setToolTip(
                '<p style="color: lime">SUCCESS!<br>Exec: {:.4f}s'.format(
                    self.logic_node.execution_time
                )
            )
            self.error_marquee.hide()

        elif self.logic_node.success in [constants.FAILED, constants.ERROR]:
            self.error_marquee.show()

            if self.logic_node.success == constants.FAILED:
                self.badge_icon.setElementId("failed")
                self.error_marquee.setPen(constants.NODE_FAILED_PEN)
                self.error_marquee.setBrush(constants.NODE_FAILED_BRUSH)

            elif self.logic_node.success == constants.ERROR:
                self.badge_icon.setElementId("error")
                self.error_marquee.setPen(constants.NODE_ERROR_PEN)
                self.error_marquee.setBrush(constants.NODE_ERROR_BRUSH)

            # Full feedback
            html_text = ""
            if self.logic_node.fail_log:
                html_text += '<p style="color: orange">FAILED:<br>' + "<br>".join(
                    self.logic_node.fail_log
                )
            if self.logic_node.error_log:
                html_text += '<p style="color: red">ERROR:<br>' + "<br>".join(
                    self.logic_node.error_log
                )

            self.badge_icon.setToolTip(html_text)

            if constants.IN_SCREEN_ERRORS:
                self.additional_info_text.setHtml(html_text)
                self.additional_info_text.show()

        # Previews
        if self.logic_node.success == constants.SUCCESSFUL:
            self.update_previews_from_attributes()

    # CHANGE ATTRIBUTES FROM INPUT WIDGETS ----------------------
    def update_attributes_from_widgets(self, *args):
        gui_internal_attr_name = args[0] if args else None
        for w in self.input_widgets:
            if (
                w.objectName() == gui_internal_attr_name
                or gui_internal_attr_name is None
            ):
                value = None

                if isinstance(w, QtWidgets.QLineEdit):
                    value = w.text() or None
                    if value is None:
                        self.logic_node[w.objectName()].clear()
                        continue
                    self.logic_node.set_attribute_from_str(w.objectName(), value)
                elif isinstance(w, QtWidgets.QPlainTextEdit):
                    value = w.toPlainText() or None
                    if value is None:
                        self.logic_node[w.objectName()].clear()
                        continue
                    self.logic_node.set_attribute_from_str(w.objectName(), value)
                elif isinstance(w, QtWidgets.QCheckBox):
                    value = w.isChecked()
                    self.logic_node[w.objectName()].set_value(value)
                elif isinstance(w, QtWidgets.QSpinBox):
                    value = w.value()
                    self.logic_node[w.objectName()].set_value(value)
                elif isinstance(w, QtWidgets.QDoubleSpinBox):
                    value = w.value()
                    self.logic_node[w.objectName()].set_value(value)
                elif isinstance(w, QtWidgets.QComboBox):
                    value = w.currentText()
                    self.logic_node[w.objectName()].set_value(value)

        if self.scene():
            GS.signals.attribute_editor_refresh_node_requested.emit(
                self.logic_node.uuid
            )

    # SHOW PREVIEWS ----------------------
    def update_previews_from_attributes(self):
        for w in self.preview_widgets:
            if not w.objectName():  # That is the label that starts the previews
                continue

            value = self.logic_node.get_attribute_value(w.objectName())

            if isinstance(w, QtWidgets.QLineEdit):
                w.setText(value)
            elif isinstance(w, QtWidgets.QPlainTextEdit):
                w.setPlainText(pprint.pformat(value))
            elif isinstance(w, QtWidgets.QCheckBox):
                w.setChecked(value)
            elif isinstance(w, QtWidgets.QSpinBox):
                w.setValue(value)
            elif isinstance(w, QtWidgets.QDoubleSpinBox):
                w.setValue(value)
            elif isinstance(w, QtWidgets.QComboBox):
                w.setCurrentText(value)
            elif isinstance(w, QtWidgets.QLabel):
                im2 = value.convert("RGB")
                data = im2.tobytes("raw", "RGB")
                image = QtGui.QImage(
                    data, im2.width, im2.height, QtGui.QImage.Format_RGB888
                )
                pix = QtGui.QPixmap.fromImage(image)
                w.setPixmap(
                    pix.scaled(w.width(), w.height(), QtCore.Qt.KeepAspectRatio)
                )

        if self.scene():
            GS.signals.attribute_editor_refresh_node_requested.emit(
                self.logic_node.uuid
            )

    def clear_previews(self):
        for w in self.preview_widgets:
            if not w.objectName():
                continue

            self.logic_node[w.objectName()].clear()

            if isinstance(w, QtWidgets.QLineEdit):
                w.setText(f"[{w.objectName()}]")
            elif isinstance(w, QtWidgets.QPlainTextEdit):
                w.setPlainText(f"[{w.objectName()}]")
            elif isinstance(w, QtWidgets.QCheckBox):
                w.setChecked(False)
            elif isinstance(w, QtWidgets.QSpinBox):
                w.setValue(0)
            elif isinstance(w, QtWidgets.QDoubleSpinBox):
                w.setValue(0.0)
            elif isinstance(w, QtWidgets.QComboBox):
                w.setCurrentIndex(0)
            elif isinstance(w, QtWidgets.QLabel):
                w.clear()
                w.setText(f"[{w.objectName()}]")

        if self.scene():
            GS.signals.attribute_editor_refresh_node_requested.emit(
                self.logic_node.uuid
            )

    # ITEM CHANGE EVENT ----------------------
    def itemChange(self, change, value):
        if change == QtWidgets.QGraphicsItem.GraphicsItemChange.ItemSelectedHasChanged:
            self.set_selected_appearance()
        elif change in [
            QtWidgets.QGraphicsItem.GraphicsItemChange.ItemScenePositionHasChanged
        ]:
            self.redraw_lines()

        return QtWidgets.QGraphicsItem.itemChange(self, change, value)

    # CONNECTIONS/LINES ----------------------
    def redraw_lines(self):
        self.scene().redraw_node_lines(self)

    def clear_all_connections(self):
        for g_a in self.graphic_attributes:
            g_a.clear_connections()

    # VISUAL FEEDBACK ----------------------
    def set_selected_appearance(self):
        self.setZValue(10)

        if self.isSelected():
            self.selection_marquee.show()
            if constants.GLOW_EFFECTS:
                self.glow.show()
        else:
            self.glow.hide()
            self.selection_marquee.hide()

    def show_help(self):
        """
        Show a window with the help/usage of this node class.
        """
        help_text = self.logic_node.get_node_html_help()
        info_window = QtWidgets.QMessageBox()
        info_window.setText(help_text)
        info_window.setWindowTitle("Help")
        info_window.setIcon(QtWidgets.QMessageBox.Information)

        f = QtCore.QFile(r"ui:stylesheet.qss")
        with open(f.fileName(), "r") as s:
            info_window.setStyleSheet(s.read())

        info_window.exec_()

    def toggle_activated(self):
        """Toggle the activated state of the node."""
        active = self.logic_node.toggle_activated()
        self.deactivated_cross.setVisible(not active)

    def show_deactivated(self):
        """Force to show deactivated."""
        self.deactivated_cross.setVisible(True)

    # UTILITY ----------------------
    def update_name(self):
        """
        Update the text shown in the node's name.
        """
        self.class_text.setHtml(
            '<p align="left"><font color=white>{0}'.format(self.logic_node.node_name)
        )

    def get_as_code(self):
        format_dict = {
            "inputs_dict": self.logic_node.INPUTS_DICT,
            "outputs_dict": self.logic_node.OUTPUTS_DICT,
            "internals_dict": self.logic_node.INTERNALS_DICT,
            "class_name": self.logic_node.node_name,
            "is_context": self.logic_node.IS_CONTEXT,
        }

        template_file = QtCore.QFile(r"resources:class_template.txt")
        with open(template_file.fileName(), "r") as template:
            template_text = template.read()
            return template_text.format(**format_dict)

    # SPECIAL METHODS ----------------------
    def __str__(self):
        return "Grapic node " + self.logic_node.node_name

    def __getitem__(self, item: str):
        for g_attr in self.graphic_attributes:
            if g_attr.logic_attribute.attribute_name == item:
                return g_attr
        LOGGER.error("Error, no graphic attribute with that name {}".format(item))


# -------------------------------- Graphic Attribute -------------------------------- #
class GeneralGraphicAttribute(QtWidgets.QGraphicsPathItem):
    def __init__(self, logic_attribute, parent_node, row_index):
        # BASIC PROPERTIES
        self.logic_attribute = logic_attribute
        self.connector_type = self.logic_attribute.connector_type
        self.row_index = row_index

        self.connected_graphic_attrs = set()

        # GRAPHIC SETUP
        QtWidgets.QGraphicsPathItem.__init__(self, parent=parent_node)
        self.setData(0, constants.GRAPHIC_ATTRIBUTE)

        self.parent_node = parent_node

        self.plug_polygon = QtWidgets.QGraphicsPathItem(parent=self)

        self.attr_text = QtWidgets.QGraphicsTextItem(parent=self)
        attr_text_font = QtGui.QFont(
            constants.NODE_FONT, int(0.4 * constants.STRIPE_HEIGHT)
        )
        self.attr_text.setFont(attr_text_font)
        self.attr_text.setHtml(
            "<font color=white>↦ {0} <font color=cyan>({1})".format(
                logic_attribute.attribute_name,
                logic_attribute.get_datatype_str(),
            )
        )
        if logic_attribute.is_optional:
            self.attr_text.setHtml(
                "<font color=#60ffffff>∗ {0} <font color=#6000FFFF>({1})".format(
                    logic_attribute.attribute_name,
                    logic_attribute.get_datatype_str(),
                )
            )
        if self.logic_attribute.connector_type == constants.OUTPUT:
            self.attr_text.setHtml(
                "<font color=cyan>({0}) <font color=white> {1} ↦".format(
                    self.logic_attribute.get_datatype_str(),
                    self.logic_attribute.attribute_name,
                )
            )
            if logic_attribute.is_optional:
                self.attr_text.setHtml(
                    "<font color=#6000FFFF>({0}) <font color=#60ffffff> {1} ∗".format(
                        logic_attribute.get_datatype_str(),
                        logic_attribute.attribute_name,
                    )
                )

        self.glow = QtWidgets.QGraphicsPathItem(parent=self)
        self.glow.hide()

    # PROPERTIES ----------------------
    @property
    def x(self):
        return self.scenePos().x()

    @property
    def center_point(self):
        return self.scenePos()

    # GEOMETRY PROPERTIES ----------------------
    def plug_coords(self):
        return self.plug_polygon.scenePos()

    # GRAPHICS SETUP ----------------------
    def setup_graphics(self):
        self.setPen(QtGui.QPen(QtCore.Qt.NoPen))

        # Plug shape (will be different shape depending on data type)
        plug_path = QtGui.QPainterPath()

        if self.logic_attribute.get_datatype_str() == "bool":
            plug_polygon = QtGui.QPolygon(
                [
                    QtCore.QPoint(-constants.PLUG_RADIUS, -constants.PLUG_RADIUS),
                    QtCore.QPoint(0, -constants.PLUG_RADIUS * 0.5),
                    QtCore.QPoint(constants.PLUG_RADIUS, -constants.PLUG_RADIUS),
                    QtCore.QPoint(constants.PLUG_RADIUS, constants.PLUG_RADIUS),
                    QtCore.QPoint(0, constants.PLUG_RADIUS * 0.5),
                    QtCore.QPoint(-constants.PLUG_RADIUS, constants.PLUG_RADIUS),
                ]
            )
            plug_path.addPolygon(plug_polygon)

        elif self.logic_attribute.get_datatype_str() == "list":
            plug_polygon = QtGui.QPolygon(
                [
                    QtCore.QPoint(-constants.PLUG_RADIUS, -constants.PLUG_RADIUS * 0.7),
                    QtCore.QPoint(-constants.PLUG_RADIUS, constants.PLUG_RADIUS * 0.7),
                    QtCore.QPoint(constants.PLUG_RADIUS, constants.PLUG_RADIUS * 0.7),
                    QtCore.QPoint(+constants.PLUG_RADIUS, -constants.PLUG_RADIUS * 0.7),
                ]
            )
            plug_path.addPolygon(plug_polygon)

        elif self.logic_attribute.get_datatype_str() == "dict":
            plug_polygon = QtGui.QPolygon(
                [
                    QtCore.QPoint(-constants.PLUG_RADIUS, 0),
                    QtCore.QPoint(0, constants.PLUG_RADIUS),
                    QtCore.QPoint(constants.PLUG_RADIUS, 0),
                    QtCore.QPoint(0, -constants.PLUG_RADIUS),
                ]
            )
            plug_path.addPolygon(plug_polygon)

        elif self.logic_attribute.get_datatype_str() == "str":
            plug_path.addEllipse(
                QtCore.QPoint(0, 0), constants.PLUG_RADIUS, constants.PLUG_RADIUS
            )

        elif self.logic_attribute.get_datatype_str() == "tuple":
            plug_polygon = QtGui.QPolygon(
                [
                    QtCore.QPoint(
                        -constants.PLUG_RADIUS * 0.8, constants.PLUG_RADIUS * 0.8
                    ),
                    QtCore.QPoint(
                        -constants.PLUG_RADIUS * 0.8, -constants.PLUG_RADIUS * 0.8
                    ),
                    QtCore.QPoint(
                        constants.PLUG_RADIUS * 0.8, constants.PLUG_RADIUS * 0.8
                    ),
                    QtCore.QPoint(
                        +constants.PLUG_RADIUS * 0.8, -constants.PLUG_RADIUS * 0.8
                    ),
                ]
            )
            plug_path.addPolygon(plug_polygon)

        elif self.logic_attribute.get_datatype_str() == "int":
            plug_polygon = QtGui.QPolygon(
                [
                    QtCore.QPoint(-constants.PLUG_RADIUS, -constants.PLUG_RADIUS),
                    QtCore.QPoint(constants.PLUG_RADIUS, -constants.PLUG_RADIUS),
                    QtCore.QPoint(constants.PLUG_RADIUS, constants.PLUG_RADIUS),
                    QtCore.QPoint(-constants.PLUG_RADIUS, constants.PLUG_RADIUS),
                ]
            )
            plug_path.addPolygon(plug_polygon)

        elif self.logic_attribute.get_datatype_str() == "float":
            plug_polygon = QtGui.QPolygon(
                [
                    QtCore.QPoint(-constants.PLUG_RADIUS, -constants.PLUG_RADIUS),
                    QtCore.QPoint(0.6 * constants.PLUG_RADIUS, -constants.PLUG_RADIUS),
                    QtCore.QPoint(constants.PLUG_RADIUS, -0.25 * constants.PLUG_RADIUS),
                    QtCore.QPoint(constants.PLUG_RADIUS, 0.25 * constants.PLUG_RADIUS),
                    QtCore.QPoint(0.6 * constants.PLUG_RADIUS, constants.PLUG_RADIUS),
                    QtCore.QPoint(-constants.PLUG_RADIUS, constants.PLUG_RADIUS),
                ]
            )
            plug_path.addPolygon(plug_polygon)

        elif self.logic_attribute.get_datatype_str() in ["Run", "RunLoop"]:
            plug_polygon = QtGui.QPolygon(
                [
                    QtCore.QPoint(-constants.PLUG_RADIUS, constants.PLUG_RADIUS),
                    QtCore.QPoint(constants.PLUG_RADIUS, 0),
                    QtCore.QPoint(-constants.PLUG_RADIUS, -constants.PLUG_RADIUS),
                ]
            )
            plug_path.addPolygon(plug_polygon)

        else:
            plug_polygon = QtGui.QPolygon()
            w = 360 / 5
            for i in range(5):
                t = w * i - 90
                x = constants.PLUG_RADIUS * math.cos(math.radians(t))
                y = constants.PLUG_RADIUS * math.sin(math.radians(t))
                plug_polygon.append(QtCore.QPoint(x, y))
            plug_path.addPolygon(plug_polygon)

        plug_path.closeSubpath()
        self.plug_polygon.setPath(plug_path)
        self.plug_polygon.setPen(
            QtGui.QPen(
                QtGui.QColor(self.parent_node.bright_color_name),
                constants.NODE_CONTOUR_THICKNESS * 0.5,
            )
        )
        self.plug_polygon.setBrush(
            QtGui.QBrush(
                QtGui.QColor(
                    self.parent_node.base_color.red() + 35,
                    self.parent_node.base_color.green() + 35,
                    self.parent_node.base_color.blue() + 35,
                )
            )
        )
        self.plug_polygon.setZValue(50)

        self.plug_polygon.setData(0, constants.PLUG)

        # Glow
        self.glow.setPath(plug_path)
        glow_pen = QtGui.QPen(
            QtGui.QColor(self.parent_node.bright_color_name),
            constants.NODE_SELECTED_GLOW_THICKNESS,
        )
        self.glow.setPen(glow_pen)
        blur = QtWidgets.QGraphicsBlurEffect()
        blur.setBlurRadius(glow_pen.width() * 1.5)
        self.glow.setGraphicsEffect(blur)
        self.glow.setZValue(40)

        # Move elements
        self.setPos(
            0,
            constants.HEADER_HEIGHT
            + self.parent_node.extra_header
            + self.row_index * constants.STRIPE_HEIGHT,
        )

        if self.logic_attribute.connector_type == constants.INPUT:
            self.plug_polygon.setPos(
                -constants.PLUG_RADIUS * 1.3,
                constants.STRIPE_HEIGHT / 2,
            )
            self.glow.setPos(
                -constants.PLUG_RADIUS * 1.3,
                constants.STRIPE_HEIGHT / 2,
            )
        elif self.logic_attribute.connector_type == constants.OUTPUT:
            self.attr_text.setPos(
                self.parent_node.node_width - self.attr_text.boundingRect().width(), 0
            )
            self.plug_polygon.setPos(
                self.parent_node.node_width + constants.PLUG_RADIUS * 1.3,
                constants.STRIPE_HEIGHT / 2,
            )
            self.glow.setPos(
                self.parent_node.node_width + constants.PLUG_RADIUS * 1.3,
                constants.STRIPE_HEIGHT / 2,
            )

    # CONNECT/DISCONNECT ----------------------
    def connect_graphic_attr(self, other_graphic_attr, check_logic=True) -> tuple:
        """
        Connect this graphic attribute to another one.

        Args:
            other_graphic_attr (GeneralGraphicAttribute): attribute to try to connect
            check_logic (bool): check the connection of logic attributes

        Returns: tuple(bool, str), whether the attribute could be connected and a message with the reason
        """
        can_connect = True
        reason = ""
        if check_logic:
            can_connect, reason = self.logic_attribute.connect_to_other(
                other_graphic_attr.logic_attribute
            )
        if can_connect:
            if self.connector_type == constants.INPUT:
                self.connected_graphic_attrs = {other_graphic_attr}
                other_graphic_attr.connected_graphic_attrs.add(self)
            else:
                self.connected_graphic_attrs.add(other_graphic_attr)
                other_graphic_attr.connected_graphic_attrs = {self}

        self.show_connected_status()
        other_graphic_attr.show_connected_status()

        return can_connect, reason

    def clear_connections(self):
        connected_g_attrs = list(self.connected_graphic_attrs)
        for connected_g_attr in connected_g_attrs:
            self.disconnect_from(connected_g_attr)

    def disconnect_from(self, other_graphic_attr):
        self.logic_attribute.disconnect_from_other(other_graphic_attr.logic_attribute)

        self.connected_graphic_attrs.remove(other_graphic_attr)
        other_graphic_attr.connected_graphic_attrs.remove(self)

        self.show_connected_status()
        other_graphic_attr.show_connected_status()

    def disconnect_input(self):
        if self.connector_type == constants.INPUT and self.connected_graphic_attrs:
            self.logic_attribute.disconnect_input()

            connected_out_graphic_attr = next(iter(self.connected_graphic_attrs))

            self.connected_graphic_attrs.remove(connected_out_graphic_attr)
            connected_out_graphic_attr.connected_graphic_attrs.remove(self)

            self.show_connected_status()
            connected_out_graphic_attr.show_connected_status()

    # UTILITY ----------------------
    def has_input_connected(self):
        return self.logic_attribute.has_input_connected()

    # VISUAL FEEDBACK ----------------------
    def show_connected_status(self):
        if self.connected_graphic_attrs:
            if constants.GLOW_EFFECTS:
                self.glow.show()
            self.plug_polygon.setPen(constants.CONNECTOR_USED_PEN)
        else:
            self.glow.hide()
            self.plug_polygon.setPen(
                QtGui.QPen(
                    QtGui.QColor(self.parent_node.bright_color_name),
                    constants.NODE_CONTOUR_THICKNESS * 0.5,
                )
            )

    # SPECIAL METHODS ----------------------
    def __str__(self):
        return self.logic_attribute.node_name



# -*- coding: UTF-8 -*-
__author__ = "Jaime Rivera <jaime.rvq@gmail.com>"
__copyright__ = "Copyright 2022, Jaime Rivera"
__credits__ = []
__license__ = "MIT License"


from all_nodes.logic.logic_node import GeneralLogicNode
from all_nodes import utils


LOGGER = utils.get_logger(__name__)


class GetEntireEnviron(GeneralLogicNode):
    NICE_NAME = "Get entire environment"

    OUTPUTS_DICT = {"environ_dict": {"type": dict}}

    def run(self):
        import os

        self.set_output("environ_dict", dict(os.environ))


class GetEnvVariable(GeneralLogicNode):
    NICE_NAME = "Get env variable"
    HELP = (
        "Get the value of an environment variable, with possibility of a fallback "
        "value if the variable is not defined"
    )

    INPUTS_DICT = {
        "env_variable_name": {"type": str},
        "fallback_value": {"type": str, "optional": True},
    }

    OUTPUTS_DICT = {"env_variable_value": {"type": str}}

    def run(self):
        import os

        env_variable_name = self.get_attribute_value("env_variable_name")
        fallback_value = self.get_attribute_value("fallback_value")
        env_var_value = os.getenv(env_variable_name)
        if env_var_value is None:
            if fallback_value:
                env_var_value = fallback_value
            else:
                self.fail(
                    "No environment variable '{}' found".format(env_variable_name)
                )
                return

        self.set_output("env_variable_value", env_var_value)


class SetEnvVariable(GeneralLogicNode):
    INPUTS_DICT = {
        "env_variable_name": {"type": str},
        "new_value": {"type": str},
    }

    def run(self):
        import os

        env_variable_name = self.get_attribute_value("env_variable_name")
        new_value = self.get_attribute_value("new_value")

        os.environ[env_variable_name] = new_value



# -*- coding: UTF-8 -*-
from __future__ import annotations

__author__ = "Jaime Rivera <jaime.rvq@gmail.com>"
__copyright__ = "Copyright 2022, Jaime Rivera"
__credits__ = []
__license__ = "MIT License"

import ast
from copy import deepcopy
import datetime
import getpass
import os
from pathlib import Path
import pprint
import re
import time
import uuid

from PySide2 import QtCore

from all_nodes import constants
from all_nodes import utils


LOGGER = utils.get_logger(__name__)


# -------------------------------- GENERAL NODE -------------------------------- #
class GeneralLogicNode:
    """
    General logic node implementation.

    Example of attribute setup:
    INPUTS_DICT = {"in_attribute_example_1": {"type": int},
                   "in_attribute_example_2": {"type": str, "optional": True},
                   "in_attribute_example_3": {"type": list, "optional": True},
                   "in_attribute_example_with_a_very_long_name": {"type": dict},
    }

    OUTPUTS_DICT = {"out_attribute_example_1": {"type": str},
                    "out_attribute_example_2": {"type": int},
    }

    The 'optional' parameter can be provided, but will always default to False unless specified,
    as we mostly want all inputs of nodes to be set.

    """

    FILEPATH = ""

    NICE_NAME = ""
    HELP = ""

    IS_CONTEXT = False
    CONTEXT_DEFINITION_FILE = ""

    INPUTS_DICT = {}
    OUTPUTS_DICT = {}

    INTERNALS_DICT = {}  # Internal attrs not exposed (for GUI input / preview mostly)

    VALID_NAMING_PATTERN = "^[A-Z]+[a-zA-Z0-9_]*$"

    def __init__(self):
        # General properties
        self.class_name = type(self).__name__
        self.node_name = self.class_name + "_1"
        self.uuid = str(uuid.uuid4())

        # Copy dicts
        self.INPUTS_DICT = deepcopy(self.INPUTS_DICT)
        self.OUTPUTS_DICT = deepcopy(self.OUTPUTS_DICT)

        # Attributes
        self.all_attributes = []
        self.check_attributes_validity()
        self.create_attributes()

        # Context it belongs to
        self.context = None

        # Specific for contexts
        self.internal_scene = None
        self.build_internal()

        # Execution
        self.active = True
        self.success = constants.NOT_RUN
        self.execution_counter = 0

        self.fail_log = []
        self.error_log = []

        self.run_date = None
        self.execution_time = 0
        self.user = getpass.getuser()

        # Signals
        self.signaler = LogicNodeSignaler()

        # Feedback
        LOGGER.debug("Initialized node! {}".format(self.class_name))

    # UTILITY ----------------------
    @staticmethod
    def name_is_valid(name) -> bool:
        """
        Check if the given name is valid according to a defined pattern.

        Parameters:
            name (str): The name to be checked.

        Returns:
            bool: True if the name is valid, False otherwise.
        """
        if re.match(GeneralLogicNode.VALID_NAMING_PATTERN, name):
            return True
        return False

    def rename(self, new_name: str) -> bool:
        """
        Rename the node to the given new name.

        Args:
            new_name (str): The new name for the node.

        Returns:
            bool: True if the node is successfully renamed, False otherwise.

        Example:
            node.rename("Node2")  # returns True
            node.rename("1Node")  # returns False
        """
        if self.node_name == new_name:
            return True

        if self.name_is_valid(new_name):
            LOGGER.debug("Renamed node '{}' to '{}'".format(self.node_name, new_name))
            self.node_name = new_name
            return True
        else:
            LOGGER.warning("Name proposed for node is not valid: {}".format(new_name))

        return False

    def force_rename(self, new_name: str) -> bool:
        """
        Rename the node to the given new name, ignoring any validation rules.

        Args:
            new_name (str): The new name for the node.

        Returns:
            bool: True always.
        """
        LOGGER.debug(
            "Forcing renaming of node '{}' to '{}'".format(self.node_name, new_name)
        )
        self.node_name = new_name
        return True

    def get_max_in_or_out_count(self) -> int:
        """
        Get the maximum number of input or output attributes.

        Returns:
            int: The maximum number of input or output attributes.
        """
        return max(len(self.get_input_attrs()), len(self.get_output_attrs()))

    def get_node_full_dict(self) -> dict:
        """
        Get a dictionary containing the full representation of the node,
        including class name, node name, attributes, connections, success, logs,
        execution info, and extra information.

        Returns:
            dict: A dictionary fully representing the node.
        """
        out_dict = dict()

        # Class name, node name...
        out_dict["class_name"] = self.class_name
        out_dict["node_name"] = self.node_name
        out_dict["full_name"] = self.full_name
        out_dict["uuid"] = self.uuid
        out_dict["origin_file"] = Path(self.FILEPATH).name

        # Attributes
        out_dict["node_attributes"] = dict()
        for attr in self.all_attributes:
            if self.INTERNALS_DICT.get(attr.attribute_name, {}).get("gui_type") in set(
                constants.PreviewsGUI
            ):
                continue  # No need to register preview attrs

            out_dict["node_attributes"][attr.dot_name] = dict()
            out_dict["node_attributes"][attr.dot_name]["attribute_name"] = (
                attr.attribute_name
            )
            out_dict["node_attributes"][attr.dot_name]["value"] = str(attr.value)
            out_dict["node_attributes"][attr.dot_name]["connector_type"] = (
                attr.connector_type
            )
            out_dict["node_attributes"][attr.dot_name]["is_optional"] = attr.is_optional
            out_dict["node_attributes"][attr.dot_name]["data_type"] = (
                attr.get_datatype_str()
            )
            out_dict["node_attributes"][attr.dot_name]["data_type_str"] = (
                attr.get_datatype_str()
            )
            out_dict["node_attributes"][attr.dot_name]["connected_to"] = [
                c_attr.dot_name for c_attr in attr.connected_attributes
            ]

        # Connections
        connections = list()
        for attr in self.all_attributes:
            connections += attr.get_connections_list()
        out_dict["connections"] = connections

        # Success and logs
        out_dict["success"] = self.success

        out_dict["error_log"] = self.error_log
        out_dict["fail_log"] = self.fail_log

        # Execution info
        out_dict["run_date"] = (
            self.run_date.strftime("%Y-%m-%d") if self.run_date else None
        )
        out_dict["execution_time"] = self.execution_time
        out_dict["user"] = self.user
        if not self.active:
            out_dict["active"] = self.active

        # Extra
        out_dict["IS_CONTEXT"] = self.IS_CONTEXT

        return out_dict

    def get_node_basic_dict(self) -> dict:
        """Get a dict that represents the main data needed to 'rebuild' a node.

        It is retrieved as dict so it can also be serialized

        Returns:
            dict: with node main data
        """
        out_dict = dict()

        out_dict[self.node_name] = dict()
        out_dict[self.node_name]["class_name"] = self.class_name
        if not self.active:
            out_dict[self.node_name]["active"] = self.active

        out_dict[self.node_name]["node_attributes"] = dict()
        for attr in self.all_attributes:
            if self.INTERNALS_DICT.get(attr.attribute_name, {}).get("gui_type") in set(
                constants.PreviewsGUI
            ):
                continue  # No need to register preview attrs

            if (
                attr.attribute_name not in [constants.START, constants.COMPLETED]
                and attr.value is not None
            ):
                out_dict[self.node_name]["node_attributes"][attr.attribute_name] = (
                    attr.value
                )

        # If no node attributes were registered, no need to keep this key
        if not out_dict[self.node_name]["node_attributes"]:
            out_dict[self.node_name].pop("node_attributes")

        return out_dict

    def get_out_connections(self):
        """
        Get a list of connections coming from output attributes of this node.

        Returns:
            list: of lists, where each list contains the dot names of two connected attributes.
        """
        connections_list = list()
        for attr in self.get_output_attrs():
            for connected in attr.connected_attributes:
                connections_list.append([attr.dot_name, connected.dot_name])

        return connections_list

    def out_connected_nodes(self):
        """
        Return a set of nodes that are connected to this node's output attributes.

        Returns:
            set: A set of nodes that are connected to this node's output attributes.
        """
        connected_nodes = set()
        for attr in self.get_output_attrs():
            for connected_attr in attr.connected_attributes:
                connected_nodes.add(connected_attr.parent_node)

        return connected_nodes

    def in_connected_nodes_recursive(self) -> list:
        """Recursively get the full 'chain' of all the nodes connected to this node's inputs

        Returns:
            list: list of nodes
        """
        in_connected_nodes = list()

        in_immediate_nodes = set()
        for attr in self.get_input_attrs():
            for connected_attr in attr.connected_attributes:
                in_immediate_nodes.add(connected_attr.parent_node)

        in_immediate_nodes_list = list(in_immediate_nodes)
        in_connected_nodes += in_immediate_nodes_list
        for node in in_immediate_nodes_list:
            in_connected_nodes += node.in_connected_nodes_recursive()

        return in_connected_nodes

    def check_cycles(self, node_to_check: GeneralLogicNode) -> bool:
        """
        Check if a given node is part of a cycle in the connected nodes of this node.

        Args:
            node_to_check (GeneralLogicNode): The node to check for cycles.

        Returns:
            bool: True if the given node is part of a cycle, False otherwise.
        """
        in_connected_nodes = self.in_connected_nodes_recursive()
        return node_to_check in in_connected_nodes

    def get_gui_internals_inputs(self) -> dict:
        """
        Return a dictionary of internal attributes that are used as GUI inputs.

        Returns:
            dict: A dictionary of internal attributes used as GUI inputs.
        """
        gui_internals_inputs = dict()
        for attr_name in self.INTERNALS_DICT:
            if self.INTERNALS_DICT[attr_name].get("gui_type") in list(
                constants.InputsGUI
            ):
                gui_internals_inputs[attr_name] = self.INTERNALS_DICT[attr_name]

        return gui_internals_inputs

    def get_gui_internals_previews(self) -> dict:
        """
        Returns a dictionary of internal attributes that are used as GUI previews.

        Returns:
            dict: A dictionary of internal attributes used as GUI previews.
        """
        gui_internals_previews = dict()
        for attr_name in self.INTERNALS_DICT:
            if self.INTERNALS_DICT[attr_name].get("gui_type") in set(
                constants.PreviewsGUI
            ):
                gui_internals_previews[attr_name] = self.INTERNALS_DICT[attr_name]

        return gui_internals_previews

    # PROPERTIES ----------------------
    @property
    def full_name(self):
        """
        Property method to get the full name of the node by combining the context full name and node name.
        Returns the full name as a string.
        """
        if self.context:
            return self.context.full_name + "/" + self.node_name
        else:
            return "/" + self.node_name

    @property
    def all_attribute_names(self):
        """
        Property method to get all attribute names by extracting attribute names from all attributes.
        Returns a list of attribute names.
        """
        return [a.attribute_name for a in self.all_attributes]

    # ATTRIBUTES ----------------------
    def check_attributes_validity(self):
        """
        Check the validity of the attributes in the node.

        Raises:
            RuntimeError: If an input attribute name is also present in the OUTPUTS_DICT.
        """
        for in_name in self.INPUTS_DICT.keys():
            if in_name in self.OUTPUTS_DICT.keys():
                raise RuntimeError(
                    "Input and output attributes cannot have same name! ({})".format(
                        in_name
                    )
                )

    def create_attributes(self):
        """
        Populate this node with the attributes that have been defined.
        """
        # -------------- INPUTS -------------- #
        # Add a special "Run control" START attribute that will be in all nodes
        start_attr = GeneralLogicAttribute(
            self, constants.START, constants.INPUT, Run, is_optional=True
        )
        self.all_attributes.append(start_attr)

        for input_attribute_name in self.INPUTS_DICT:
            in_attr = GeneralLogicAttribute(
                self,
                input_attribute_name,
                constants.INPUT,
                self.INPUTS_DICT[input_attribute_name]["type"],
                is_optional=self.INPUTS_DICT[input_attribute_name].get(
                    "optional", False
                ),
            )
            self.all_attributes.append(in_attr)

        # -------------- OUTPUTS -------------- #
        for output_attribute_name in self.OUTPUTS_DICT:
            out_attr = GeneralLogicAttribute(
                self,
                output_attribute_name,
                constants.OUTPUT,
                self.OUTPUTS_DICT[output_attribute_name]["type"],
                is_optional=self.OUTPUTS_DICT[output_attribute_name].get(
                    "optional", False
                ),
            )
            self.all_attributes.append(out_attr)

        # Add a special "Run control" COMPLETED attribute that will be in all nodes
        completed_attr = GeneralLogicAttribute(
            self, constants.COMPLETED, constants.OUTPUT, Run, is_optional=True
        )
        self.all_attributes.append(completed_attr)

        # -------------- INTERNAL -------------- #
        for internal_attribute_name in self.INTERNALS_DICT:
            internal_attr = GeneralLogicAttribute(
                self,
                internal_attribute_name,
                constants.INTERNAL,
                self.INTERNALS_DICT[internal_attribute_name]["type"],
                is_optional=self.INTERNALS_DICT[internal_attribute_name].get(
                    "optional", False
                ),
            )
            self.all_attributes.append(internal_attr)

    def add_attribute(
        self,
        attribute_name,
        connector_type,
        data_type,
        gui_type=None,
        is_optional=False,
        value=None,
    ):
        # Check name
        if attribute_name in self.all_attribute_names:
            LOGGER.error(
                "Attribute name '{}' is already used. Skipping.".format(attribute_name)
            )
            return

        # Create attribute
        new_attribute = GeneralLogicAttribute(
            self, attribute_name, connector_type, data_type, is_optional=is_optional
        )

        if value:
            new_attribute.set_value(value)

        self.all_attributes.append(new_attribute)

        # Registrer it to the dict of the instance
        if connector_type == constants.INPUT:
            self.INPUTS_DICT[attribute_name] = dict()
            self.INPUTS_DICT[attribute_name]["type"] = data_type
            self.INPUTS_DICT[attribute_name]["optional"] = is_optional
        elif connector_type == constants.OUTPUT:
            self.OUTPUTS_DICT[attribute_name] = dict()
            self.OUTPUTS_DICT[attribute_name]["type"] = data_type
            self.OUTPUTS_DICT[attribute_name]["optional"] = is_optional
        elif connector_type == constants.INTERNAL:
            self.INTERNALS_DICT[attribute_name] = dict()
            self.INTERNALS_DICT[attribute_name]["type"] = data_type
            self.INTERNALS_DICT[attribute_name]["gui_type"] = gui_type

        # Return new attribute
        return new_attribute

    def get_input_attrs(self):
        """
        Get all the input attributes of the node.

        Returns:
            list: with all the input attributesf
        """
        return [
            attr
            for attr in self.all_attributes
            if attr.connector_type == constants.INPUT
        ]

    def clear_input_attrs(self):
        for attr in self.get_input_attrs():
            attr.clear()

    def recursive_clear_connected_input_attrs(self):
        for attr in self.get_input_attrs():
            if attr.has_input_connected():
                attr.clear()

        for node in self.out_connected_nodes():
            node.recursive_clear_connected_input_attrs()

    def recursive_set_in_loop(self):
        self.success = constants.IN_LOOP

        for node in self.out_connected_nodes():
            if node.class_name == "ForEachEnd":  # TODO cleanup this
                continue
            node.recursive_set_in_loop()

    def get_output_attrs(self):
        """
        Get all the output attributes of the node.

        Returns:
            list: with all the output attributes
        """
        return [
            attr
            for attr in self.all_attributes
            if attr.connector_type == constants.OUTPUT
        ]

    def set_attribute_value(self, attribute_name: str, value):
        """
        Set an attribute to a new value.

        Args:
            attribute_name (str): name of the attribute to set
            value: new value to set the attribute to
        """
        if attribute_name not in self.all_attribute_names:
            raise RuntimeError(
                "Error! No valid attribute '{}' in the node {}".format(
                    attribute_name, self.node_name
                )
            )

        for attribute in self.all_attributes:
            if attribute.attribute_name == attribute_name:
                if isinstance(value, attribute.data_type):
                    attribute.set_value(value)
                else:
                    raise RuntimeError(
                        "Not a valid type! {} not valid for {} (needed: {})".format(
                            value,
                            attribute.dot_name,
                            attribute.get_datatype_str(),
                        )
                    )

    def set_input(self, attribute_name: str, value):
        """
        Set an input attribute to a new value.

        Args:
            attribute_name (str): name of the input attribute to set
            value: new value to set the attribute to
        """
        for attr in self.all_attributes:
            if (
                attr.attribute_name == attribute_name
                and attr.connector_type == constants.INPUT
            ):
                self.set_attribute_value(attribute_name, value)
                return

        LOGGER.error(
            "Error! No valid input attribute {} in the node".format(attribute_name)
        )

    def set_output(self, attribute_name: str, value):
        """
        Set an output attribute to a new value.

        Args:
            attribute_name (str): name of the output attribute to set
            value: new value to set the attribute to
        """
        for attr in self.all_attributes:
            if (
                attr.attribute_name == attribute_name
                and attr.connector_type == constants.OUTPUT
            ):
                self.set_attribute_value(attribute_name, value)
                return

        LOGGER.error(
            "Error! No valid output attribute {} in the node".format(attribute_name)
        )

    def set_attribute_from_str(self, attribute_name: str, value_str: str):
        """
        Set the value of an attribute from a string.

        Args:
            attribute_name (str): The name of the attribute to set.
            value_str (str): The string value to set the attribute to.

        Notes:
            - If the attribute name is not valid, an error message is logged and the function returns.
            - If the value string is empty, the attribute is cleared.
        """
        if attribute_name not in self.all_attribute_names:
            LOGGER.error(
                "Error! No valid attribute '{}' in the node {}".format(
                    attribute_name, self.node_name
                )
            )
            return

        if value_str == "":
            for attribute in self.all_attributes:
                if attribute.attribute_name == attribute_name:
                    attribute.clear()
                    return

        for attribute in self.all_attributes:
            if attribute.attribute_name == attribute_name:
                if attribute.data_type is str:
                    attribute.set_value(value_str)
                elif attribute.data_type is float:
                    attribute.set_value(float(value_str))
                elif attribute.data_type is int:
                    attribute.set_value(int(value_str))
                elif attribute.data_type is bool:
                    if value_str in ["0", "1"]:
                        attribute.set_value(bool(int(value_str)))
                    elif value_str.lower() in ["false", "true"]:
                        attribute.set_value(
                            {
                                "False": False,
                                "True": True,
                                "false": False,
                                "true": True,
                            }.get(value_str)
                        )
                elif attribute.data_type in (dict, list, object, tuple):
                    try:
                        attribute.set_value(ast.literal_eval(value_str))
                    except (SyntaxError, ValueError) as e:
                        attribute.clear()
                        LOGGER.debug(e)
                else:
                    LOGGER.error(
                        "Cannot set value {} to type {}, not defined how to cast from string".format(
                            value_str, attribute.data_type
                        )
                    )

    def get_attribute_value(self, attribute_name: str):
        """
        Get the value of an attribute with the given name.

        Args:
            attribute_name (str): The name of the attribute to retrieve the value of.

        Returns:
            The value of the attribute, or None if the attribute does not exist.
        """
        if attribute_name not in self.all_attribute_names:
            LOGGER.error(
                "Error! No valid attribute '{}' in the node {}".format(
                    attribute_name, self.node_name
                )
            )
            return

        return self[attribute_name].get_value()

    def connect_attribute(
        self,
        attribute_name: str,
        other_node: GeneralLogicNode,
        other_attribute_name: str,
    ):
        """
        Connect the given attribute of this node to the given attribute of another node.

        Args:
            attribute_name (str): The name of the attribute in this node.
            other_node (GeneralLogicNode): The other node to connect to.
            other_attribute_name (str): The name of the attribute in the other node.

        Returns:
            The result of the connection between the source attribute and the target attribute.
        """
        source_attr = self[attribute_name]
        target_attr = other_node[other_attribute_name]
        return source_attr.connect_to_other(target_attr)

    # CHECKS ----------------------
    def is_starting_node(self) -> bool:
        """
        Determine whether or not this node can be a starting point for execution.

        Returns:
            bool
        """
        # If the node has some inputs connected, it is not a starting point
        for attr in self.all_attributes:
            if attr.has_input_connected():
                return False

        # Then, count all non-optional inputs
        needed_input_attrs_count = 0
        for attr in self.all_attributes:
            if attr.connector_type == constants.INPUT and not attr.is_optional:
                if attr.is_empty():
                    needed_input_attrs_count += 1

        if needed_input_attrs_count == 0:
            return True

        return False

    def check_all_inputs_have_value(self) -> bool:
        """
        Check if all needed input attributes have a value.

        Returns:
            bool: True if all needed input attributes have a value, False otherwise.
        """
        not_set_attrs = []
        for attr in self.get_input_attrs():
            if attr.is_optional and attr.is_empty():
                if attr.has_input_connected():
                    not_set_attrs.append(attr.dot_name)
            elif attr.is_empty():
                not_set_attrs.append(attr.dot_name)

        if not_set_attrs:
            LOGGER.warning(
                "Some input attributes of node {} not set: {}".format(
                    self.full_name, not_set_attrs
                )
            )
            return False

        return True

    def check_all_outputs_have_value(self) -> bool:
        """
        Check if all needed output attributes have a value.

        Returns:
            bool: True if all output attributes have a value, False otherwise.
        """
        not_set_attrs = []
        for attr in self.get_output_attrs():
            if attr.is_empty() and not attr.is_optional:
                not_set_attrs.append(attr.dot_name)

        if not_set_attrs:
            LOGGER.warning(
                "Some output attributes of node {} not set: {}".format(
                    self.full_name, not_set_attrs
                )
            )
            return False

        return True

    # CONTEXT ----------------------
    def is_in_root(self):
        """
        Check if the node is in the root context.

        Returns:
            bool
        """
        return self.context is None

    def build_internal(self):
        """
        Build the internal scene for the context node.

        If the node is not a context node, the function returns without doing anything.
        """
        if not self.IS_CONTEXT:
            return

        from all_nodes.logic.logic_scene import LogicScene

        LOGGER.debug(
            "Building internal scene for context node {}".format(self.full_name)
        )

        self.internal_scene = LogicScene()
        self.internal_scene.context = self
        self.internal_scene.set_name(self.full_name)

        context_definition_file = os.path.join(
            os.path.dirname(os.path.abspath(self.FILEPATH)),
            self.class_name + ".ctx",
        )
        if os.path.isfile(context_definition_file):
            self.CONTEXT_DEFINITION_FILE = context_definition_file
            self.internal_scene.load_from_file(self.CONTEXT_DEFINITION_FILE)
            self.internal_scene.set_context_to_nodes()
        else:
            raise RuntimeError(
                "No context definition file found for {}. Expected at: {}".format(
                    self.class_name, context_definition_file
                )
            )

    def set_context(self, context: GeneralLogicNode):
        """
        Set the context for this node.

        Args:
            context (GeneralLogicNode): The context to set this node as part of
        """
        self.context = context

    # RUN ----------------------
    def _run(self, execute_connected=True):
        """
        Run the node.

        Parameters:
            execute_connected (bool): Whether to execute connected nodes. Default is True
        """
        # ------------------- PRE-CHECKS ------------------- #
        # --------------- Status
        if self.success not in [constants.NOT_RUN, constants.IN_LOOP]:
            LOGGER.warning(
                "Cannot execute node {} it has already been executed!".format(
                    self.full_name
                )
            )
            self.signaler.finished.emit()
            return

        # --------------- If inactive, lets just skip it
        if not self.active:
            LOGGER.warning(
                "Skipping execution of {}".format(
                    self.full_name,
                )
            )

            self.success = constants.NOT_RUN
            self.set_output(constants.COMPLETED, Run())
            self.propagate_results()

            self.signaler.finished.emit()

            # Execute connected, if they can (might be missing inputs from this one)
            if execute_connected:
                for node in self.out_connected_nodes():
                    LOGGER.info(
                        "From {} (skipped), launching execution of {}".format(
                            self.full_name, node.full_name
                        )
                    )
                    node._run()
                return

        # --------------- Check inputs
        if not self.check_all_inputs_have_value():
            LOGGER.debug(
                "Cannot execute node {} now, some input attributes are not set".format(
                    self.full_name
                )
            )
            self.signaler.finished.emit()
            return

        # ------------------- START EXECUTION ------------------- #
        LOGGER.info(
            "Starting execution of {} ({})".format(self.full_name, self.class_name)
        )
        self.execution_counter += 1

        t1 = time.time()
        self.run_date = datetime.datetime.now()
        self.signaler.is_executing.emit()

        # --------------- Clear any previous logging
        self.fail_log = []
        self.error_log = []

        # --------------- Run
        if self.IS_CONTEXT:
            self.internal_scene.run_all_nodes(
                spawn_thread=False
            )  # TODO maybe this can be improved? / recursive
            internal_failures = self.internal_scene.gather_failed_nodes_logs()
            if internal_failures:
                for f in internal_failures:
                    self.fail(f)
            internal_errors = self.internal_scene.gather_errored_nodes_logs()
            if internal_errors:
                for e in internal_errors:
                    self.error(e)

        else:
            try:
                self.run()
            except Exception as e:
                self.error(str(e))
                LOGGER.exception(e)
                self.signaler.finished.emit()
                self.execution_time = time.time() - t1
                return

        # --------------- Result of Run
        if self.success == constants.FAILED:
            LOGGER.error(
                "Execution of {} FAILED, cannot keep executing from this node".format(
                    self.full_name
                )
            )
            self.signaler.finished.emit()
            self.execution_time = time.time() - t1
            return

        elif self.success == constants.ERROR:
            LOGGER.error(
                "Execution of {} ERRORED, cannot keep executing from this node".format(
                    self.full_name
                )
            )
            self.signaler.finished.emit()
            self.execution_time = time.time() - t1
            return

        # --------------- Check outputs were all set during Run
        if not self.check_all_outputs_have_value():
            LOGGER.error(
                "Something went wrong, not all output attributes are set in {}, "
                "cannot keep executing from this node".format(self.full_name)
            )
            self.fail("Not all output attributes are set")
            self.execution_time = time.time() - t1
            return

        # --------------- Mark successful
        self.success = constants.SUCCESSFUL
        self.set_output(constants.COMPLETED, Run())

        # --------------- Propagate results
        LOGGER.debug(
            "From {}, propagating out attributes to connected nodes".format(
                self.full_name
            )
        )
        self.propagate_results()

        # --------------- Stop timer and emit signal
        self.execution_time = time.time() - t1
        self.signaler.finished.emit()

        # --------------- Execute connected
        if execute_connected:
            for node in self.out_connected_nodes():
                if node.success not in [constants.FAILED, constants.ERROR]:
                    LOGGER.info(
                        "From {}, launching execution of {}".format(
                            self.full_name, node.full_name
                        )
                    )
                    node._run()

    def run_single(self):
        """
        Run only this node.
        """
        self._run(execute_connected=False)

    def run_chain(self):
        """
        Run this node and then all the nodes connected to its outputs.
        """
        self._run(execute_connected=True)

    def run(self):
        """
        To be reimplemented in each subclass.
        """
        if self.IS_CONTEXT:
            LOGGER.debug("Contexts do not need to implement the 'run' method")
        else:
            self.fail(
                "Class '{}' does not have the 'run' method implemented!".format(
                    self.class_name
                )
            )

    def fail(self, message=None):
        """
        Mark this node as failed.

        Args:
            message (str): Optional message to display when failed
        """
        self.success = constants.FAILED
        if message:
            LOGGER.error(
                "{} FAILED: {}. Cannot keep executing from this node".format(
                    self.full_name, message
                )
            )
            self.fail_log.append(message)

    def error(self, message=None):
        """
        Mark this node as errored.

        Args:
            message (str): Optional message to display when errored
        """
        self.success = constants.ERROR
        if message:
            LOGGER.error(
                "{} ERRORED: {}. Cannot keep executing from this node".format(
                    self.full_name, message
                )
            )
            self.error_log.append(message)

    def mark_skipped(self):
        """
        Mark this node as skipped.
        """
        self.success = constants.SKIPPED
        self.signaler.finished.emit()

    def propagate_results(self):
        """
        Propagate the values of calculated output attrs to the other connected nodes input attrs.
        """
        for attr in self.get_output_attrs():
            attr.propagate_value()

    def reset(self):
        """
        Reset this node.

        Marks the node as not-executed, and all information that could have been left by a
        previous execution (input values, logs...).
        """
        LOGGER.info("Resetting node " + self.full_name)
        for attr in self.get_input_attrs():
            if attr.has_connections():
                attr.clear()
        for attr in self.get_output_attrs():
            attr.clear()

        self.success = constants.NOT_RUN
        self.fail_log = []
        self.error_log = []
        self.execution_time = 0

        if self.IS_CONTEXT:
            self.internal_scene.reset_all_nodes()

    def soft_reset(self):
        """
        Perform a 'soft reset'

        Mainly for debugging purposes. Will reset the status of the node so it can be run, but without
        deleting the input values of it.
        """
        LOGGER.info("Soft-resetting node " + self.full_name)

        self.success = constants.NOT_RUN
        self.fail_log = []
        self.error_log = []
        self.execution_time = 0

        if self.IS_CONTEXT:
            self.internal_scene.soft_reset_all_nodes()

    def toggle_activated(self):
        """Toggle the activated state of the node."""
        self.active = not self.active
        return self.active

    # SPECIAL METHODS ----------------------
    def __getitem__(self, item: str):
        for attr in self.all_attributes:
            if attr.attribute_name == item:
                return attr
        LOGGER.error(
            "Error, no attribute with that name {}.{}".format(self.full_name, item)
        )
        return None

    def __str__(self):
        return "<{} object>".format(
            self.full_name,
        )

    def __repr__(self):
        return "<{} object>".format(
            self.full_name,
        )

    def get_node_html_help(self):
        """
        Get the node's help documentation parsed into an html text.

        Returns:
            str: with the complete text of the help
        """
        help_text = f"<h1>{self.class_name}</h1>"
        help_text += (
            f"<p style='white-space:pre'>Class defined at:<br>{self.FILEPATH}</p>"
        )

        if self.HELP:
            help_text += f"<p style='white-space:pre'>HELP:<br>{self.HELP}"

        help_text += (
            "<h4><br>INPUTS_DICT:</h4><p style='white-space:pre; background-color:black'><code>"
            + pprint.pformat(self.INPUTS_DICT, indent=2)
            .replace(">", "&gt;")
            .replace("<", "&lt;")
            .replace("\n", "<br>")
            + "</code></p>"
        )

        help_text += (
            "<h4>OUTPUTS_DICT:</h4><p style='white-space:pre; background-color:black'><code>"
            + pprint.pformat(self.OUTPUTS_DICT, indent=2)
            .replace(">", "&gt;")
            .replace("<", "&lt;")
            .replace("\n", "<br>")
            + "</code></p>"
        )

        help_text += (
            "<h4>INTERNALS_DICT:</h4><p style='white-space:pre; background-color:black'><code>"
            + pprint.pformat(self.INTERNALS_DICT, indent=2)
            .replace(">", "&gt;")
            .replace("<", "&lt;")
            .replace("\n", "<br>")
            + "</code></p>"
        )

        return help_text


class LogicNodeSignaler(QtCore.QObject):
    """
    Defines the signals available from a logic node.
    """

    status_changed = QtCore.Signal()
    is_executing = QtCore.Signal()
    finished = QtCore.Signal()


# -------------------------------- ATTRIBUTE -------------------------------- #
class GeneralLogicAttribute:
    def __init__(
        self,
        parent_node: GeneralLogicNode,
        attribute_name: str,
        connector_type: str,
        data_type,
        value=None,
        is_optional=False,
    ):
        self.parent_node = parent_node

        self.attribute_name = attribute_name
        self.connector_type = connector_type  # IN or OUT
        self.data_type = data_type
        self.value = value
        self.is_optional = is_optional

        self.connected_attributes = set()

    # PROPERTIES ----------------------
    @property
    def dot_name(self):
        return self.parent_node.node_name + "." + self.attribute_name

    @property
    def full_name(self):
        return self.parent_node.full_name + "." + self.attribute_name

    # GET AND SET ----------------------
    def get_value(self):
        return self.value

    def is_empty(self) -> bool:
        """
        Check if the value of the attribute is None.

        Returns:
            bool
        """
        return self.value is None

    def set_value(self, new_value):
        """
        Set the value of the attribute to the given new value.

        Args:
            new_value (Any): The new value to set the object to.
        """
        self.value = new_value
        LOGGER.debug("Setting {} to new value {}".format(self.full_name, self.value))

    def clear(self):
        """
        Clear the value of the attribute.

        This method sets the value of the attribute to None, effectively clearing its current value.
        """
        self.value = None

    # CONNECTIONS ----------------------
    def get_connections_list(self):
        """
        Get a list of connections where this attribute is involved.

        Returns:
            list: A list of lists, where each list contains the dot names of two connected attributes.
        """
        connections = []
        if self.has_connections():
            for connected_attr in self.connected_attributes:
                if self.connector_type == constants.INPUT:
                    connections.append([connected_attr.dot_name, self.dot_name])
                else:
                    connections.append([self.dot_name, connected_attr.dot_name])

        return connections

    def has_input_connected(self) -> bool:
        """Determine whether or not this attribute has incoming connections.

        Returns:
            bool: whether or not the attribute has any incoming connection
        """
        return (
            self.connector_type == constants.INPUT
            and len(self.connected_attributes) > 0
        )

    def has_connections(self) -> bool:
        """Determine whether or not this attribute is connected to another.

        Returns:
            bool: whether or not the attribute has any connection
        """
        return len(self.connected_attributes) > 0

    def connect_to_other(self, other_attribute: GeneralLogicAttribute) -> tuple:
        """
        Connect this attribute to another.

        Args:
            other_attribute (GeneralLogicAttribute)

        Returns:
            tuple(bool, str): whether or not the connection could be done, log of the reason
        """
        # Checks -------------------------
        # Cycles check
        if self.connector_type == constants.INPUT:
            if other_attribute.parent_node.check_cycles(self.parent_node):
                connection_warning = "Cannot connect, cycle detected!"
                LOGGER.warning(connection_warning)
                return (False, connection_warning)
        elif self.connector_type == constants.OUTPUT:
            if self.parent_node.check_cycles(other_attribute.parent_node):
                connection_warning = "Cannot connect, cycle detected!"
                LOGGER.warning(connection_warning)
                return (False, connection_warning)

        # Different nodes check
        if not self.parent_node != other_attribute.parent_node:
            connection_warning = "Cannot connect {} and {}, both same node {}".format(
                self.dot_name,
                other_attribute.dot_name,
                self.parent_node.full_name,
            )
            LOGGER.warning(connection_warning)
            return (False, connection_warning)

        # IN - OUT check
        if not self.connector_type != other_attribute.connector_type:
            connection_warning = "Cannot connect {} and {}, both are {}".format(
                self.dot_name, other_attribute.dot_name, self.connector_type
            )
            LOGGER.warning(connection_warning)
            return (False, "Cannot connect both are {}".format(self.connector_type))

        # Datatype check
        connection_direction = "->"
        if self.connector_type == constants.INPUT:
            connection_direction = "<-"
        if self.data_type != other_attribute.data_type:
            if object not in [self.data_type, other_attribute.data_type]:
                connection_warning = (
                    "Cannot connect {} {} {}, different datatypes {}{}{}".format(
                        self.dot_name,
                        connection_direction,
                        other_attribute.dot_name,
                        self.get_datatype_str(),
                        connection_direction,
                        other_attribute.get_datatype_str(),
                    )
                )
                LOGGER.warning(connection_warning)
                return (
                    False,
                    "Cannot connect, different datatypes {}{}{}".format(
                        self.get_datatype_str(),
                        connection_direction,
                        other_attribute.get_datatype_str(),
                    ),
                )

        # Connection -------------------------
        if self.connector_type == constants.OUTPUT:
            self.connected_attributes.add(other_attribute)
            other_attribute.connected_attributes = {self}
        else:
            self.connected_attributes = {other_attribute}
            other_attribute.connected_attributes.add(self)

        connection_log = "Connected {} {} {}".format(
            self.dot_name, connection_direction, other_attribute.dot_name
        )
        LOGGER.info(connection_log)

        return (True, connection_log)

    def disconnect_from_other(self, other_attribute: GeneralLogicAttribute):
        """
        Disconnect this attribute from another.

        Args:
            other_attribute (GeneralLogicAttribute)
        """
        self.connected_attributes.remove(other_attribute)
        other_attribute.connected_attributes.remove(self)
        if self.connector_type == constants.OUTPUT:
            LOGGER.info(
                "Disconnected {} -/- {}".format(self.dot_name, other_attribute.dot_name)
            )
        else:
            LOGGER.info(
                "Disconnected {} -/- {}".format(other_attribute.dot_name, self.dot_name)
            )

    def disconnect_input(self):
        """
        Disconnect incoming inputs to this attribute.
        """
        if self.connector_type == constants.INPUT and self.connected_attributes:
            self.disconnect_from_other(next(iter(self.connected_attributes)))

    # UTILITY ----------------------
    def get_datatype_str(self) -> str:
        """
        Get a string representation of the datatype that this attribute can hold.

        Returns:
            str: with a representation fo the datatype
        """
        return utils.parse_datatype(str(self.data_type))

    def propagate_value(self):
        """
        Propagate the value of this attribute to all its connected attributes.
        """
        for connected_attr in self.connected_attributes:
            connected_attr.set_value(self.value)

    # SPECIAL METHODS ----------------------
    def __str__(self):
        return "GeneralLogicAttribute: {}, value:{}".format(self.dot_name, self.value)


# -------------------------------- UTILITY -------------------------------- #
class Run:
    """
    Utility class with no functionality
    """

    def __str__(self):
        return "Run"


class RunLoop:
    """
    Utility class with no functionality
    """

    def __str__(self):
        return "RunLoop"



__author__ = "Jaime Rivera <jaime.rvq@gmail.com>"
__copyright__ = "Copyright 2022, Jaime Rivera"
__credits__ = []
__license__ = "MIT License"


import datetime

from all_nodes.constants import InputsGUI
from all_nodes.logic.logic_node import GeneralLogicNode
from all_nodes import utils


LOGGER = utils.get_logger(__name__)


DATE_FORMATS = sorted(
    [
        "%d/%m/%Y,  %H:%M:%S",  # European
        "%m/%d/%Y, %H:%M:%S",  # American
        "%Y.%m.%d_%H.%M.%S",  # Technical
    ]
)


class StrfDatetime(GeneralLogicNode):
    HELP = "Format a datetime object into a string"

    INPUTS_DICT = {
        "datetime_object": {"type": datetime.datetime},
        "datetime_formatting": {"type": str},
    }
    OUTPUTS_DICT = {"datetime_formatted": {"type": str}}

    def run(self):
        datetime_object = self.get_attribute_value("datetime_object")
        datetime_formatting = self.get_attribute_value("datetime_formatting")
        self.set_output(
            "datetime_formatted", datetime_object.strftime(datetime_formatting)
        )


class DatetimeFormatsSelect(GeneralLogicNode):
    NICE_NAME = "Datetime formats"

    INTERNALS_DICT = {
        "internal_datetime_format_str": {
            "type": str,
            "gui_type": InputsGUI.OPTION_INPUT,
            "options": DATE_FORMATS,
        },
    }

    OUTPUTS_DICT = {"out_datetime_format_str": {"type": str}}

    def run(self):
        self.set_output(
            "out_datetime_format_str",
            self.get_attribute_value("internal_datetime_format_str"),
        )


class DatetimeNow(GeneralLogicNode):
    NICE_NAME = "Datetime now"
    HELP = "Get a datetime object as of right now, as well as a formatted string"

    INTERNALS_DICT = {
        "internal_datetime_str": {
            "type": str,
            "gui_type": InputsGUI.OPTION_INPUT,
            "options": DATE_FORMATS,
        },
    }

    OUTPUTS_DICT = {
        "datetime_object": {"type": datetime.datetime},
        "out_datetime_str": {"type": str},
    }

    def run(self):
        datetime_object = datetime.datetime.now()
        datetime_formatting = self.get_attribute_value("internal_datetime_str")
        self.set_output(
            "out_datetime_str", datetime_object.strftime(datetime_formatting)
        )
        self.set_output("datetime_object", datetime_object)



__author__ = "Jaime Rivera <jaime.rvq@gmail.com>"
__copyright__ = "Copyright 2022, Jaime Rivera"
__credits__ = []
__license__ = "MIT License"


import os
import tempfile
import unittest

from all_nodes import constants
from all_nodes.lib.base_node_lib.nodes_general_library import debug
from all_nodes.lib.base_node_lib.nodes_general_library import file_reading
from all_nodes.lib.base_node_lib.nodes_general_library import file_writing
from all_nodes.lib.base_node_lib.nodes_general_library import folder_management
from all_nodes.lib.base_node_lib.nodes_general_library import miscellaneous
from all_nodes.lib.base_node_lib.nodes_general_library import general_input
from all_nodes import utils


# -------------------------------- TESTS -------------------------------- #
class NodeTesting(unittest.TestCase):
    DICT_EXAMPLE = {
        "test_people": {"names": ["Fred", "Joe", "Sally"], "surnames": None}
    }

    def test_creation(self):
        """
        Create a few nodes, directly from their modules.
        """
        utils.print_test_header("test_creation")

        n_1 = debug.EmptyNode()
        self.assertIsNotNone(n_1)
        n_2 = miscellaneous.GetDictKey()
        self.assertIsNotNone(n_2)
        n_3 = folder_management.CopyFoldersRecursive()
        self.assertIsNotNone(n_3)

    def test_attribute_name(self):
        """
        Test the attributes of a node get created with a proper name.
        """
        utils.print_test_header("test_attribute_name")

        n_1 = miscellaneous.GetDictKey()
        in_dict_attr = n_1["in_dict"]
        self.assertRegex(in_dict_attr.dot_name, "GetDictKey_\d+\.in_dict")

    def test_connection(self):
        """
        Test connecting some node attributes.
        """
        utils.print_test_header("test_connection")

        n_1 = file_reading.JsonToDict()
        n_2 = miscellaneous.GetDictKey()
        self.assertTrue(n_1.connect_attribute("out_dict", n_2, "in_dict")[0])
        self.assertFalse(n_1["out_dict"].connect_to_other(n_2["key"])[0])
        self.assertFalse(n_2["in_dict"].connect_to_other(n_2["key"])[0])
        self.assertFalse(n_2["key"].connect_to_other(n_1["out_dict"])[0])

    def test_connection_cycle(self):
        """
        Test connecting some node attributes.
        """
        utils.print_test_header("test_connection_cycle")

        n_1 = debug.EmptyNode()
        n_2 = debug.EmptyNode()
        n_3 = debug.EmptyNode()
        n_1.connect_attribute(constants.COMPLETED, n_2, constants.START)
        n_2.connect_attribute(constants.COMPLETED, n_3, constants.START)
        connection_status, _ = n_3.connect_attribute(
            constants.COMPLETED, n_1, constants.START
        )
        self.assertFalse(connection_status)

    def test_getting_internal_dict(self):
        """
        Test displaying the internal debug dict of a node
        """
        utils.print_test_header("test_getting_internal_dict")

        n_s = general_input.StrInput()
        n_d = miscellaneous.GetDictKey()

        n_s.set_attribute_value("internal_str", "test")
        n_s.connect_attribute("out_str", n_d, "key")
        n_d.set_attribute_value("in_dict", {"test": 100})
        n_s.run_chain()
        assert n_d["out"].get_value() == 100

        assert n_d.get_node_full_dict()

    def test_inputs_checked_run(self):
        """
        Check all inputs of a node are properly set before starting to execute.
        """
        utils.print_test_header("test_inputs_checked_run")

        n_e = debug.EmptyNode()
        n_e.run_single()
        self.assertEqual(n_e.success, constants.SUCCESSFUL)

        n_j = file_reading.JsonToDict()
        n_j.run_single()
        self.assertEqual(n_j.success, constants.NOT_RUN)

        n_f = file_writing.CreateTempFile()
        n_f.run_single()
        self.assertEqual(n_f.success, constants.SUCCESSFUL)

        n_f2 = file_writing.CreateTempFile()
        n_f2["suffix"].set_value(".txt")
        n_f2.run_single()
        self.assertEqual(n_f2.success, constants.SUCCESSFUL)

    def test_run_node_1(self):
        """
        Test the DictToYaml node
        """
        utils.print_test_header("test_run_node_1")

        temp = tempfile.NamedTemporaryFile(suffix=".yaml", delete=False)
        temp.close()
        n_1 = file_writing.DictToYaml()
        n_1.set_attribute_value("in_dict", NodeTesting.DICT_EXAMPLE)
        n_1.set_attribute_value("yaml_filepath_to_write", temp.name)
        n_1._run()
        self.assertTrue(os.path.exists(temp.name))
        os.unlink(temp.name)
        self.assertFalse(os.path.exists(temp.name))

    def test_run_node_2(self):
        """
        Test the CreateTempFile node
        """
        utils.print_test_header("test_run_node_2")

        n_1 = file_writing.CreateTempFile()
        n_1._run()
        out_temp_path = n_1.get_attribute_value("tempfile_path")
        self.assertTrue(os.path.exists(out_temp_path))
        os.unlink(out_temp_path)
        self.assertFalse(os.path.exists(out_temp_path))

    def test_node_fails(self):
        """
        Test a case where a node should fail
        """
        utils.print_test_header("test_node_fails")

        n_1 = miscellaneous.GetDictKey()
        n_1.set_attribute_value("in_dict", NodeTesting.DICT_EXAMPLE)
        n_1.set_attribute_value("key", "FAKE")
        n_1._run()

        self.assertEqual(n_1.success, constants.FAILED)

    def test_node_exception(self):
        """
        Test a case where a node should error and display the error
        """
        utils.print_test_header("test_node_exception")

        n_1 = debug.ErrorNode()
        n_1._run()

        self.assertEqual(n_1.success, constants.ERROR)

    def test_node_execution_time(self):
        """
        Test the execution time of nodes is being measured
        """
        utils.print_test_header("test_node_execution_time")

        n_1 = debug.TimedNode()
        n_1.run_single()

        self.assertGreater(n_1.execution_time, 1.0)

    def test_add_attrs(self):
        """
        Check attributes can be added to a node instance
        """
        utils.print_test_header("test_add_attrs")

        n_empty = debug.EmptyNode()
        n_empty.add_attribute("test_attr", constants.INPUT, int, value=100)
        self.assertEqual(n_empty["test_attr"].get_value(), 100)
        self.assertEqual(len(n_empty.get_input_attrs()), 2)

    def test_add_attrs_2(self):
        """
        Check attributes can be added to a node instance
        """
        utils.print_test_header("test_add_attrs_2")

        n_empty = debug.ErrorNode()
        n_empty.add_attribute("some_attr", constants.OUTPUT, str)
        self.assertIsNone(n_empty["some_attr"].get_value())
        self.assertEqual(len(n_empty.get_output_attrs()), 2)



# -*- coding: UTF-8 -*-
__author__ = "Jaime Rivera <jaime.rvq@gmail.com>"
__copyright__ = "Copyright 2022, Jaime Rivera"
__credits__ = []
__license__ = "MIT License"


from all_nodes.constants import InputsGUI
from all_nodes.logic.logic_node import GeneralLogicNode
from all_nodes import utils


LOGGER = utils.get_logger(__name__)


class StrInput(GeneralLogicNode):
    NICE_NAME = "String input"

    OUTPUTS_DICT = {"out_str": {"type": str}}

    INTERNALS_DICT = {
        "internal_str": {"type": str, "gui_type": InputsGUI.STR_INPUT},
    }

    def run(self):
        self.set_output("out_str", self.get_attribute_value("internal_str"))


class MultilineStrInput(GeneralLogicNode):
    NICE_NAME = "Multiline string input"
    HELP = ""

    OUTPUTS_DICT = {"out_str": {"type": str}}

    INTERNALS_DICT = {
        "internal_str": {"type": str, "gui_type": InputsGUI.MULTILINE_STR_INPUT},
    }

    def run(self):
        self.set_output("out_str", self.get_attribute_value("internal_str"))


class BoolInput(GeneralLogicNode):
    NICE_NAME = "Boolean input"
    HELP = ""

    OUTPUTS_DICT = {"out_bool": {"type": bool}}

    INTERNALS_DICT = {
        "internal_bool": {"type": bool, "gui_type": InputsGUI.BOOL_INPUT},
    }

    def run(self):
        self.set_output("out_bool", self.get_attribute_value("internal_bool"))


class DictInput(GeneralLogicNode):
    NICE_NAME = "Dictionary input"
    HELP = ""

    OUTPUTS_DICT = {"out_dict": {"type": dict}}

    INTERNALS_DICT = {
        "internal_dict": {"type": dict, "gui_type": InputsGUI.DICT_INPUT},
    }

    def run(self):
        internal_dict = self.get_attribute_value("internal_dict")
        if internal_dict is None:
            self.fail(
                "Looks like the dictionary was not properly formed! (Check the syntax)"
            )
        self.set_output("out_dict", self.get_attribute_value("internal_dict"))


class ListInput(GeneralLogicNode):
    NICE_NAME = "List input"
    HELP = ""

    OUTPUTS_DICT = {"out_list": {"type": list}}

    INTERNALS_DICT = {
        "internal_list": {"type": list, "gui_type": InputsGUI.LIST_INPUT},
    }

    def run(self):
        internal_list = self.get_attribute_value("internal_list")
        if internal_list is None:
            self.fail("Looks like the list was not properly formed! (Check the syntax)")
        self.set_output("out_list", self.get_attribute_value("internal_list"))


class TupleInput(GeneralLogicNode):
    NICE_NAME = "Tuple input"
    HELP = ""

    OUTPUTS_DICT = {"out_tuple": {"type": tuple}}

    INTERNALS_DICT = {
        "internal_tuple": {"type": tuple, "gui_type": InputsGUI.TUPLE_INPUT},
    }

    def run(self):
        internal_tuple = self.get_attribute_value("internal_tuple")
        if internal_tuple is None:
            self.fail(
                "Looks like the tuple was not properly formed! (Check the syntax)"
            )
        self.set_output("out_tuple", self.get_attribute_value("internal_tuple"))


class IntInput(GeneralLogicNode):
    NICE_NAME = "Integer input"
    HELP = ""

    OUTPUTS_DICT = {"out_int": {"type": int}}

    INTERNALS_DICT = {
        "internal_int": {"type": int, "gui_type": InputsGUI.INT_INPUT},
    }

    def run(self):
        self.set_output("out_int", self.get_attribute_value("internal_int"))


class FloatInput(GeneralLogicNode):
    NICE_NAME = "Float input"
    HELP = ""

    OUTPUTS_DICT = {"out_float": {"type": float}}

    INTERNALS_DICT = {
        "internal_float": {"type": float, "gui_type": InputsGUI.FLOAT_INPUT},
    }

    def run(self):
        self.set_output("out_float", self.get_attribute_value("internal_float"))


class OptionInput(GeneralLogicNode):
    OUTPUTS_DICT = {"out_str": {"type": str}}

    INTERNALS_DICT = {
        "internal_str": {
            "type": str,
            "gui_type": InputsGUI.OPTION_INPUT,
            "options": ["1", "2", "3"],
        },
    }

    def run(self):
        self.set_output("out_str", self.get_attribute_value("internal_str"))



__author__ = "Jaime Rivera <jaime.rvq@gmail.com>"
__copyright__ = "Copyright 2022, Jaime Rivera"
__credits__ = []
__license__ = "MIT License"


import os
import unittest
import tempfile

from all_nodes import constants
from all_nodes.logic.logic_scene import LogicScene
from all_nodes.logic.logic_scene import LogicSceneError
from all_nodes import utils


# -------------------------------- TESTS -------------------------------- #
class LogicSceneTesting(unittest.TestCase):
    DICT_EXAMPLE = {
        "test_people": {"names": ["Fred", "Joe", "Sally"], "surnames": None}
    }

    FIXTURES_FOLDER = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "fixtures"
    )

    def test_logic_scene(self):
        """
        Create nodes in a logic scene
        """
        utils.print_test_header("test_logic_scene")

        logic_scene = LogicScene()
        logic_scene.add_node_by_name("JsonToDict")
        logic_scene.add_node_by_name("GetDictKey")
        logic_scene.add_node_by_name("GetDictKey")
        logic_scene.add_node_by_name("FailNode")
        self.assertEqual(logic_scene.node_count(), 4)

    def test_run_node_graph_starting_at_node(self):
        utils.print_test_header("test_run_node_graph_starting_at_node")

        logic_scene = LogicScene()
        n_1 = logic_scene.add_node_by_name("StrInput")
        n_1.set_attribute_value("internal_str", "test_people")
        n_2 = logic_scene.add_node_by_name("GetDictKey")
        n_2.set_attribute_value("in_dict", LogicSceneTesting.DICT_EXAMPLE)
        n_2["key"].connect_to_other(n_1["out_str"])
        n_3 = logic_scene.add_node_by_name("PrintToConsole")
        n_3["in_object_0"].connect_to_other(n_2["out"])
        n_1.run_chain()
        self.assertEqual(n_1.success, constants.SUCCESSFUL)
        self.assertEqual(n_2.success, constants.SUCCESSFUL)

    def test_locate_starting_nodes(self):
        utils.print_test_header("test_locate_starting_nodes")

        logic_scene = LogicScene()
        logic_scene.add_node_by_name("StrInput")
        n_2 = logic_scene.add_node_by_name("GetDictKey")
        logic_scene.add_node_by_name("PrintToConsole")
        logic_scene.add_node_by_name("CreateTempFile")
        n_5 = logic_scene.add_node_by_name("CreateTempFile")
        n_5[constants.START].connect_to_other(n_2[constants.COMPLETED])
        starting_nodes = logic_scene.get_starting_nodes()
        self.assertEqual(len(starting_nodes), 3)

    def test_locate_starting_nodes_2(self):
        utils.print_test_header("test_locate_starting_nodes_2")

        logic_scene = LogicScene()
        n_1 = logic_scene.add_node_by_name("GetEnvVariable")
        n_1["env_variable_name"].set_value("USER")
        n_2 = logic_scene.add_node_by_name("PrintToConsole")
        n_1["env_variable_value"].connect_to_other(n_2["in_object_0"])
        starting_nodes = logic_scene.get_starting_nodes()
        self.assertEqual(len(starting_nodes), 1)

    def test_run_scene(self):
        utils.print_test_header("test_run_scene")

        logic_scene = LogicScene()
        n_1 = logic_scene.add_node_by_name("GetEnvVariable")
        n_1["env_variable_name"].set_value("USER")
        n_1["fallback_value"].set_value("NO_USER_FOUND")
        n_2 = logic_scene.add_node_by_name("PrintToConsole")
        n_1["env_variable_value"].connect_to_other(n_2["in_object_0"])
        logic_scene.run_all_nodes()
        self.assertTrue(n_1.success)
        self.assertTrue(n_2.success)

    def test_capture_node(self):
        utils.print_test_header("test_capture_node")

        logic_scene = LogicScene()
        n_1 = logic_scene.add_node_by_name("StrInput")
        full_name = n_1.node_name
        n_2 = logic_scene.to_node(full_name)
        self.assertEqual(n_1, n_2)

    def test_run_scene_with_disabled(self):
        utils.print_test_header("test_run_scene_with_disabled")

        logic_scene = LogicScene()
        logic_scene.load_from_file("fail_scene")
        logic_scene.run_all_nodes()

        n = logic_scene.to_node("EmptyNode_2")
        assert n.success == constants.NOT_RUN
        assert not n.active

    def test_write_scene_to_file(self):
        utils.print_test_header("test_write_scene_to_file")

        logic_scene = LogicScene()
        n_1 = logic_scene.add_node_by_name("StrInput")
        n_1.set_attribute_value("internal_str", "TEST")
        n_2 = logic_scene.add_node_by_name("PrintToConsole")
        n_1["out_str"].connect_to_other(n_2["in_object_0"])
        n_2[constants.START].connect_to_other(n_1[constants.COMPLETED])
        logic_scene.run_all_nodes()

        temp = tempfile.NamedTemporaryFile(suffix=".yaml", delete=False)
        temp.close()
        logic_scene.save_to_file(temp.name)

        self.assertTrue(os.path.isfile(temp.name))

    def test_load_scene_from_file(self):
        utils.print_test_header("test_load_scene_from_file")

        logic_scene = LogicScene()
        logic_scene.load_from_file(
            os.path.join(self.FIXTURES_FOLDER, "environ_to_yaml_and_json.yml")
        )
        self.assertEqual(logic_scene.node_count(), 9)

    def test_load_scene_from_file_faulty(self):
        utils.print_test_header("test_load_scene_from_file_2")

        logic_scene = LogicScene()
        with self.assertRaises(LogicSceneError) as e:
            logic_scene.load_from_file(
                os.path.join(self.FIXTURES_FOLDER, "faulty_scene.yml")
            )
        print(e.exception)

    def test_load_scene_from_file_faulty_2(self):
        utils.print_test_header("test_load_scene_from_file_3")

        logic_scene = LogicScene()
        with self.assertRaises(LogicSceneError) as e:
            logic_scene.load_from_file(
                os.path.join(self.FIXTURES_FOLDER, "faulty_scene_2.yml")
            )
        print(e.exception)

    def test_load_scene_from_alias(self):
        utils.print_test_header("test_load_scene_from_alias")

        logic_scene = LogicScene()
        logic_scene.load_from_file("simple_regex")

    def test_load_scene_from_alias_faulty(self):
        utils.print_test_header("test_load_scene_from_alias_faulty")

        logic_scene = LogicScene()
        with self.assertRaises(LogicSceneError) as e:
            logic_scene.load_from_file("fake_scene")
        print(e.exception)

    def test_execute_scene_from_alias(self):
        utils.print_test_header("test_execute_scene_from_alias")

        logic_scene = LogicScene()
        logic_scene.load_from_file("simple_regex")
        logic_scene.run_all_nodes_batch()

        assert len(logic_scene.gather_failed_nodes_logs()) == 0
        assert len(logic_scene.gather_errored_nodes_logs()) == 0

    def test_execute_scene_from_alias_2(self):
        utils.print_test_header("test_execute_scene_from_alias_2")

        logic_scene = LogicScene()
        logic_scene.load_from_file("environ_to_yaml")
        logic_scene.run_all_nodes_batch()

        assert len(logic_scene.gather_failed_nodes_logs()) == 0
        assert len(logic_scene.gather_errored_nodes_logs()) == 0

    def test_execute_scene_from_alias_3(self):
        utils.print_test_header("test_execute_scene_from_alias_3")

        logic_scene = LogicScene()
        logic_scene.load_from_file("fail_scene")
        logic_scene.run_all_nodes_batch()

        assert len(logic_scene.gather_failed_nodes_logs()) == 3
        assert len(logic_scene.gather_errored_nodes_logs()) == 3

    def test_execute_scene_from_alias_4(self):
        utils.print_test_header("test_execute_scene_from_alias_4")

        logic_scene = LogicScene()
        logic_scene.load_from_file("datetime_example")
        logic_scene.run_all_nodes_batch()

        assert len(logic_scene.gather_failed_nodes_logs()) == 0
        assert len(logic_scene.gather_errored_nodes_logs()) == 0

    def test_rename_node_incorrect(self):
        utils.print_test_header("test_rename_node_incorrect")

        logic_scene = LogicScene()
        n = logic_scene.add_node_by_name("PrintToConsole")
        result = logic_scene.rename_node(n, "Bad_name_*?")
        self.assertFalse(result)

    def test_rename_node_already_present(self):
        utils.print_test_header("test_rename_node_already_present")

        logic_scene = LogicScene()
        n_1 = logic_scene.add_node_by_name("PrintToConsole")
        n_2 = logic_scene.add_node_by_name("YamlToDict")
        self.assertFalse(logic_scene.rename_node(n_1, n_2.node_name))

    def test_scene_with_loop(self):
        utils.print_test_header("test_scene_with_loop")

        # The loops are quite tricky, lets run this test a
        # few times to make sure it works, accounting for
        # randomization in execution order of nodes
        for _ in range(5):
            logic_scene = LogicScene()
            logic_scene.load_from_file("loop_example")
            logic_scene.run_all_nodes_batch()

            foreach_begin = logic_scene.to_node("ForEachBegin_1")
            assert foreach_begin.execution_counter == 1

            print_1 = logic_scene.to_node("PrintToConsole_1")
            assert print_1.execution_counter == 3

            print_2 = logic_scene.to_node("PrintToConsole_2")
            assert print_2.execution_counter == 3

            foreach_end = logic_scene.to_node("ForEachEnd_1")
            assert foreach_end.execution_counter == 1

            empty_node_2 = logic_scene.to_node("EmptyNode_2")
            assert empty_node_2.execution_counter == 1



# -*- coding: UTF-8 -*-
__author__ = "Jaime Rivera <jaime.rvq@gmail.com>"
__copyright__ = "Copyright 2022, Jaime Rivera"
__credits__ = []
__license__ = "MIT License"


from PySide2 import QtGui
from PySide2 import QtWidgets
from PySide2 import QtCore

from all_nodes.graphic.widgets.global_signaler import GlobalSignaler


GS = GlobalSignaler()


class FeedbackLineEdit(QtWidgets.QLineEdit):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setAcceptDrops(True)
        self.setReadOnly(True)
        self.setFont(QtGui.QFont("arial", 13))
        self.hide()

    def dragEnterEvent(self, event):
        event.acceptProposedAction()

    def dropEvent(self, event):
        event.ignore()
        GS.signals.dropped_node.emit(
            self.mapToParent(self.rect().topLeft())
            + event.pos()
            - QtCore.QPoint(20, 20)
        )

    def mousePressEvent(self, event):
        event.ignore()
        super().mousePressEvent(event)



# -*- coding: UTF-8 -*-
__author__ = "Jaime Rivera <jaime.rvq@gmail.com>"
__copyright__ = "Copyright 2022, Jaime Rivera"
__credits__ = []
__license__ = "MIT License"


from PySide2 import QtCore
from PySide2 import QtGui
from PySide2 import QtWidgets

from all_nodes import utils
from all_nodes.graphic.widgets.global_signaler import GlobalSignaler
from all_nodes.logic.class_registry import CLASS_REGISTRY as CR


GS = GlobalSignaler()

LOGGER = utils.get_logger(__name__)


class ClassSearcher(QtWidgets.QWidget):
    """TODO this description"""

    def __init__(self, *args, **kwargs):
        QtWidgets.QWidget.__init__(self, *args, **kwargs)

        self.layout = QtWidgets.QVBoxLayout()
        self.setLayout(self.layout)

        self.search_bar = QtWidgets.QLineEdit()
        self.search_bar.setPlaceholderText("Search for a class...")
        self.layout.addWidget(self.search_bar)

        self.class_list = QtWidgets.QListWidget()
        self.layout.addWidget(self.class_list)

        self.make_connections()

    def make_connections(self):
        self.search_bar.textChanged.connect(self.filter_classes)
        self.class_list.itemClicked.connect(self.send_node_creation_signal)

    def reset(self):
        """
        Reset the state of the widget.
        """
        self.search_bar.clear()
        self.class_list.clear()
        self.show()  # TODO should this be separate?
        self.search_bar.setFocus()

    def filter_classes(self):
        """
        Filter the list of classes based on the text entered in the search bar.
        """
        self.class_list.clear()
        current_text = self.search_bar.text()
        if not current_text:
            return
        for class_name, icon_path in CR.get_all_classes_simplified():
            if current_text.lower() in class_name.lower():
                i = QtWidgets.QListWidgetItem(class_name)
                i.setIcon(QtGui.QIcon(icon_path))
                self.class_list.addItem(i)

        if self.class_list.item(0):
            self.class_list.item(0).setSelected(True)

    def send_node_creation_signal(self):
        """
        Send a signal to request the creation of a node based on selection
        """
        if self.class_list.selectedItems():
            GS.signals.node_creation_requested.emit(
                self.pos(),
                self.class_list.selectedItems()[0].text(),
            )
        self.hide()

    def keyPressEvent(self, event: QtWidgets.QWidget.event):
        QtWidgets.QWidget.keyPressEvent(self, event)

        modifiers = QtWidgets.QApplication.keyboardModifiers()

        if event.key() == QtCore.Qt.Key_Return and not modifiers:
            self.send_node_creation_signal()

        elif event.key() == QtCore.Qt.Key_Down and not modifiers:
            index = self.class_list.indexFromItem(self.class_list.selectedItems()[0])
            row = index.row()
            new_selected = self.class_list.item(row + 1)
            if new_selected:
                self.class_list.selectedItems()[0].setSelected(False)
                new_selected.setSelected(True)

        elif event.key() == QtCore.Qt.Key_Up and not modifiers:
            index = self.class_list.indexFromItem(self.class_list.selectedItems()[0])
            row = index.row()
            new_selected = self.class_list.item(row - 1)
            if new_selected:
                self.class_list.selectedItems()[0].setSelected(False)
                new_selected.setSelected(True)

        event.setAccepted(True)



__author__ = "Jaime Rivera <jaime.rvq@gmail.com>"
__copyright__ = "Copyright 2022, Jaime Rivera"
__credits__ = []
__license__ = "MIT License"


from all_nodes.logic.logic_node import GeneralLogicNode
from all_nodes import utils
import random


LOGGER = utils.get_logger(__name__)


class RandomRGBA(GeneralLogicNode):
    NICE_NAME = "Random RGBA"
    HELP = "Generate a random RGBA color"

    OUTPUTS_DICT = {"out_rgba": {"type": tuple}}

    def run(self):
        rgba = (random.random(), random.random(), random.random(), random.random())
        self.set_output("out_rgba", rgba)


class RandomInt(GeneralLogicNode):
    NICE_NAME = "Random Int"
    HELP = "Generate a random integer"

    INPUTS_DICT = {
        "min": {"type": int, "optional": True},
        "max": {"type": int, "optional": True},
    }

    OUTPUTS_DICT = {"out_int": {"type": int}}

    def run(self):
        min_val = self.get_attribute_value("min") or 0
        max_val = self.get_attribute_value("max") or 100
        random_int = random.randint(min_val, max_val)
        self.set_output("out_int", random_int)


class RandomFloat(GeneralLogicNode):
    NICE_NAME = "Random Float"
    HELP = "Generate a random float"

    INPUTS_DICT = {
        "min": {"type": float, "optional": True},
        "max": {"type": float, "optional": True},
    }

    OUTPUTS_DICT = {"out_float": {"type": float}}

    def run(self):
        min_val = self.get_attribute_value("min") or 0.0
        max_val = self.get_attribute_value("max") or 1.0
        random_float = random.uniform(min_val, max_val)
        self.set_output("out_float", random_float)


class RandomBool(GeneralLogicNode):
    NICE_NAME = "Random Bool"
    HELP = "Generate a random boolean value"

    OUTPUTS_DICT = {"out_bool": {"type": bool}}

    def run(self):
        random_bool = random.choice([True, False])
        self.set_output("out_bool", random_bool)



# -*- coding: UTF-8 -*-
__author__ = "Jaime Rivera <jaime.rvq@gmail.com>"
__copyright__ = "Copyright 2022, Jaime Rivera"
__credits__ = []
__license__ = "MIT License"


import os
from functools import partial
from pathlib import Path


from PySide2 import QtCore
from PySide2 import QtGui
from PySide2 import QtUiTools
from PySide2 import QtWidgets

from all_nodes import constants
from all_nodes.graphic import graphic_scene
from all_nodes.graphic.widgets.attribute_editor import AttributeEditor
from all_nodes.graphic.widgets.global_signaler import GlobalSignaler
from all_nodes.graphic.widgets.shortcuts_help import ShortcutsHelp
from all_nodes.logic.class_registry import CLASS_REGISTRY as CR
from all_nodes import utils


GS = GlobalSignaler()

LOGGER = utils.get_logger(__name__)


class AllNodesWindow(QtWidgets.QMainWindow):
    def __init__(self):
        QtWidgets.QMainWindow.__init__(self)

        # CLASSES SCANNING
        self.libraries_added = set()

        # Add search paths
        root_dir_path = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        QtCore.QDir.addSearchPath("icons", os.path.join(root_dir_path, "general_icons"))
        QtCore.QDir.addSearchPath(
            "graphics", os.path.join(root_dir_path, "general_graphics")
        )
        QtCore.QDir.addSearchPath("ui", os.path.join(root_dir_path, "ui"))
        QtCore.QDir.addSearchPath(
            "resources", os.path.join(root_dir_path, "../logic/resources")
        )

        # Load UI
        file = QtCore.QFile(r"ui:all_nodes.ui")
        file.open(QtCore.QFile.ReadOnly)
        loader = QtUiTools.QUiLoader()
        self.ui = loader.load(file, self)
        self.setCentralWidget(self.ui)

        # Dock
        self.attr_editr_dock = QtWidgets.QDockWidget()
        self.attr_editr_dock.setWindowTitle("Attribute Editor")
        self.attr_editor = AttributeEditor()
        self.attr_editr_dock.setWidget(self.attr_editor)

        # STYLE
        self.setMinimumWidth(850)
        self.setMinimumHeight(500)
        self.resize(1550, 850)
        self.setWindowTitle("ALL NODES")
        self.setWindowIcon(QtGui.QIcon("icons:nodes_2.png"))

        # FEEDBACK
        self.the_process_window = QtWidgets.QTextEdit()
        self.the_process_window.setReadOnly(True)
        self.the_process_window.setWindowTitle("Execution feedback")

        # ELEMENTS OF THE UI
        self.ui.nodes_tree.setMinimumWidth(300)
        self.ui.nodes_tree.setDragEnabled(True)
        self.ui.annotations_tree.setDragEnabled(True)

        self.ui.splitter_libs.setStretchFactor(0, 8)
        self.ui.splitter_libs.setStretchFactor(1, 2)

        self.add_scene()
        self.create_dock_windows()

        # STYLESHEET
        f = QtCore.QFile(r"ui:stylesheet.qss")
        with open(f.fileName(), "r") as s:
            self.setStyleSheet(s.read())

        # SET CONNECTIONS
        self.make_connections()

        # INITIALIZE
        self.show()
        LOGGER.debug("all_nodes main window created")

        # MENUS
        self.create_menus()

        # INITIAL STATE OF WIDGETS (while scanning for classes)
        # We dont want the user to load any scene until the scanning is done, so we deactivate the menubar
        self.menuBar().setDisabled(True)

    def make_connections(self):
        """
        Establish all connections between widget signals and methods.
        """
        # UI elements
        self.ui.filter_le.textChanged.connect(self.filter_nodes_by_name)

        self.ui.tabWidget.currentChanged.connect(self.show_scene_results)

        self.ui.nodes_tree.itemEntered.connect(self.ui.annotations_tree.clearSelection)
        self.ui.annotations_tree.itemEntered.connect(self.ui.nodes_tree.clearSelection)

        # Classes scanning / populating
        for worker in CR.get_workers():
            worker.signaler.finished.connect(self.populate_tree)

        GS.signals.class_scanning_finished.connect(
            lambda: self.menuBar().setEnabled(True)
        )

        GS.signals.class_scanning_finished.connect(self.populate_annotations)

        # Global signaler
        GS.signals.dropped_node.connect(self.add_node_to_current)
        GS.signals.node_creation_requested.connect(self.add_node_to_current)

        GS.signals.tab_names_refresh_requested.connect(self.refresh_tab_names)

        GS.signals.attribute_editor_node_addition_requested.connect(
            self.add_node_to_attribute_editor_by_uuid
        )
        GS.signals.attribute_editor_refresh_node_requested.connect(
            self.refresh_node_in_attribute_editor_by_uuid
        )
        GS.signals.attribute_editor_remove_node_requested.connect(
            self.remove_node_in_attribute_editor_by_uuid
        )

        GS.signals.context_expansion_requested.connect(self.expand_context)

        GS.signals.attribute_editor_global_refresh_requested.connect(
            self.attr_editor.refresh
        )

    # UI SETUP ----------------------
    def create_menus(self):
        """
        Create and populate menu bar.
        """
        all_scenes = CR.get_all_scenes()

        def add_scenes_recursive(entries_dict: dict, menu: QtWidgets.QMenu):
            """
            Navigate a dict with multiple levels and add scenes to the menu accordingly.

            Args:
                entries_dict (dict): containing the scenes to be added
                menu (QtWidgets.QMenu): menu to add entries to
            """
            menu.setToolTipsVisible(True)
            for key in entries_dict:
                nice_name = key.replace("scene_lib", "").title().replace("_", " ")
                libs_menu = menu.addMenu(nice_name)
                libs_menu.setIcon(QtGui.QIcon("icons:folder.svg"))
                libs_menu.setToolTipsVisible(True)
                scenes_list = entries_dict[key]
                for elem in scenes_list:
                    if isinstance(elem, dict):
                        add_scenes_recursive(elem, libs_menu)
                for elem in scenes_list:
                    if isinstance(elem, tuple):
                        scene_name, full_path = elem
                        nice_name = scene_name.title().replace("_", " ")
                        ac = QtWidgets.QAction(nice_name, parent=menu)
                        ac.setToolTip(full_path)
                        ac.triggered.connect(partial(self.load_scene, full_path))
                        ac.setIcon(QtGui.QIcon("icons:scene.png"))
                        libs_menu.addAction(ac)

        menu = self.menuBar()

        file_menu = menu.addMenu("&File")
        new_scene_action = QtWidgets.QAction("Clear workspace", self)
        new_scene_action.setIcon(
            QtGui.QIcon("icons:clear.png")
        )  # TODO investigate if extensions can be removed
        new_scene_action.triggered.connect(self.clear_workspace)
        file_menu.addAction(new_scene_action)
        save_scene_action = QtWidgets.QAction("Save current scene as", self)
        save_scene_action.setIcon(QtGui.QIcon("icons:save_as.svg"))
        save_scene_action.triggered.connect(self.save_scene_as)
        file_menu.addAction(save_scene_action)
        save_scene_changes_action = QtWidgets.QAction("Save changes", self)
        save_scene_changes_action.setIcon(QtGui.QIcon("icons:save.svg"))
        save_scene_changes_action.triggered.connect(self.save_scene)
        file_menu.addAction(save_scene_changes_action)
        load_scene_action = QtWidgets.QAction("Load scene from file", self)
        load_scene_action.setIcon(QtGui.QIcon("icons:load.png"))
        load_scene_action.triggered.connect(self.load_scene)
        file_menu.addAction(load_scene_action)

        node_menu = menu.addMenu("&Node library")
        re_scan_classes_action = QtWidgets.QAction("Re-scan classes", self)
        re_scan_classes_action.setIcon(QtGui.QIcon("icons:examine.png"))
        re_scan_classes_action.triggered.connect(self.reload_classes)
        node_menu.addAction(re_scan_classes_action)

        scene_menu = menu.addMenu("&Scene library")
        add_scenes_recursive(all_scenes, scene_menu)

        window_menu = menu.addMenu("&Window")
        show_attr_editor = QtWidgets.QAction("Show attribute editor", self)
        show_attr_editor.setIcon(QtGui.QIcon("icons:eye.svg"))
        show_attr_editor.triggered.connect(self.attr_editr_dock.show)
        window_menu.addAction(show_attr_editor)

        help_menu = menu.addMenu("&Help")
        shortcuts_action = QtWidgets.QWidgetAction(help_menu)
        shortcuts_action.setDefaultWidget(ShortcutsHelp())
        help_menu.addAction(shortcuts_action)

    def create_dock_windows(self):
        self.attr_editr_dock.setAllowedAreas(
            QtCore.Qt.LeftDockWidgetArea | QtCore.Qt.RightDockWidgetArea
        )
        self.addDockWidget(QtCore.Qt.RightDockWidgetArea, self.attr_editr_dock)
        self.attr_editr_dock.hide()

    def populate_annotations(self):
        annotation_items_folder = QtCore.QDir.searchPaths("graphics")[0]

        for note_type_elem in os.listdir(annotation_items_folder):
            note_type = Path(note_type_elem).stem
            annotation_item = QtWidgets.QTreeWidgetItem()
            annotation_item.setIcon(0, QtGui.QIcon(f"graphics:{note_type}.svg"))
            annotation_item.setText(0, note_type.replace("_", " ").title())
            annotation_item.setData(0, QtCore.Qt.UserRole, note_type)
            annotation_item.setData(
                0, QtCore.Qt.UserRole + 1, constants.GRAPHIC_ANNOTATION
            )
            annotation_item.setFont(0, QtGui.QFont("arial", 14))
            self.ui.annotations_tree.addTopLevelItem(annotation_item)

        self.ui.annotations_tree.sortByColumn(0, QtCore.Qt.AscendingOrder)

    def populate_tree(self):
        """
        Populate the tree where all nodes are displayed.
        """
        lib_names = CR.get_all_classes().keys()
        for lib_name in lib_names:
            if lib_name in self.libraries_added:
                continue
            self.libraries_added.add(lib_name)

            all_classes = CR.get_all_classes().get(lib_name)
            if not all_classes:
                return

            top_level_items = {}
            for lib in sorted(all_classes):
                node_lib_path = all_classes[lib]["node_lib_path"]
                node_lib_name = all_classes[lib]["node_lib_name"]
                node_lib_nice_name = node_lib_name.title().replace("_", " ")
                module_filename = all_classes[lib]["module_filename"]
                module_full_path = all_classes[lib]["module_full_path"]
                module_nice_name = lib.title().replace("_", " ")
                color = all_classes[lib].get("color", constants.DEFAULT_NODE_COLOR)

                if node_lib_name not in top_level_items:
                    lib_item = QtWidgets.QTreeWidgetItem()
                    lib_item.setText(0, node_lib_nice_name)
                    lib_item.setToolTip(0, "Lib: {}".format(node_lib_path))
                    top_level_items[node_lib_name] = lib_item
                else:
                    lib_item = top_level_items[node_lib_name]

                module_item = QtWidgets.QTreeWidgetItem()
                module_item.setText(0, module_nice_name)
                module_item.setToolTip(0, "Module: {}".format(module_filename))

                for name, cls in all_classes[lib]["classes"]:
                    class_item = QtWidgets.QTreeWidgetItem()
                    class_item.setText(0, name)
                    class_item.setToolTip(
                        0,
                        "<h2>Class:&nbsp;<b>{}</b></h2><p style='white-space:pre'>{}".format(
                            name, module_full_path
                        ),
                    )
                    class_item.setData(0, QtCore.Qt.UserRole, name)
                    class_item.setData(
                        0, QtCore.Qt.UserRole + 1, constants.GRAPHIC_NODE
                    )
                    if (
                        hasattr(cls, "NICE_NAME") and cls.NICE_NAME
                    ):  # TODO inheritance not working here?
                        class_item.setText(0, cls.NICE_NAME)

                    icon = QtGui.QIcon(cls.ICON_PATH)
                    class_item.setIcon(0, icon)
                    item_color = QtGui.QColor(color)
                    item_color.setAlphaF(0.25)
                    class_item.setBackgroundColor(0, item_color)
                    module_item.addChild(class_item)

                module_item_color = QtGui.QColor(color)
                module_item_color.setAlphaF(0.4)
                module_item.setBackgroundColor(0, module_item_color)

                lib_item.addChild(module_item)

            for t in top_level_items:
                lib_item = top_level_items[t]
                self.ui.nodes_tree.addTopLevelItem(lib_item)
                lib_item.setExpanded(True)
                lib_item.setFont(0, QtGui.QFont("arial", 16))
                for i in range(lib_item.childCount()):
                    node_file_item = lib_item.child(i)
                    node_file_item.setExpanded(True)
                    node_file_item.setFont(0, QtGui.QFont("arial", 14))
                    for i in range(node_file_item.childCount()):
                        node_class = node_file_item.child(i)
                        node_class.setFont(0, QtGui.QFont("arial", 14))

            self.ui.nodes_tree.sortByColumn(0, QtCore.Qt.AscendingOrder)

    def filter_nodes_by_name(self):
        """
        Set node classes as hidden/visible based on user input.
        """
        filter = self.ui.filter_le.text().lower()
        for i in range(self.ui.nodes_tree.topLevelItemCount()):
            lib_item = self.ui.nodes_tree.topLevelItem(i)
            lib_visible_count = 0
            lib_item.setHidden(False)
            for m_i in range(lib_item.childCount()):
                module_item = lib_item.child(m_i)
                module_visible_count = 0
                module_item.setHidden(False)
                for c_i in range(module_item.childCount()):
                    class_item = module_item.child(c_i)
                    if filter and filter not in class_item.text(0).lower():
                        class_item.setHidden(True)
                    else:
                        class_item.setHidden(False)
                        lib_visible_count += 1
                        module_visible_count += 1
                if not module_visible_count:
                    module_item.setHidden(True)
            if not lib_visible_count:
                lib_item.setHidden(True)

    def reload_classes(self):
        """
        Reload the classes displayed in the nodes tree widget.

        This function clears the nodes tree widget and the libraries added list.
        It then flushes the class registry and starts the classes scanning.
        """
        # Clear the UI
        self.menuBar().setEnabled(False)
        self.ui.nodes_tree.clear()
        self.ui.annotations_tree.clear()
        self.libraries_added.clear()

        # Re-scan classes
        CR.flush()
        CR.scan_for_classes_GUI()

        for worker in CR.get_workers():
            worker.signaler.finished.connect(self.populate_tree)

    # TABS ----------------------
    def clear_workspace(self):
        """
        Clear the workspace by displaying a confirmation dialog to the user.

        If the user confirms, all unsaved progress will be lost. It clears the tabs and adds a new scene to the workspace.
        """
        dlg = QtWidgets.QMessageBox(parent=None)
        dlg.setWindowTitle("Clear workspace")
        dlg.setText(
            "<p style='white-space:pre'>Are you sure?<br>All unsaved progress will be lost!"
        )
        dlg.setStandardButtons(QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
        dlg.setIcon(QtWidgets.QMessageBox.Question)
        button = dlg.exec_()

        if button == QtWidgets.QMessageBox.No:
            return

        self.clear_tabs()
        self.add_scene()

    def clear_tabs(self):
        """
        Remove all tabs / scenes
        """
        self.ui.tabWidget.clear()

    def add_scene(self, context=None):
        """
        Add a new graphic scene to the widget (in a new tab).

        If a context is supplied, build its internals onto the new scene.

        Args:
            context (GeneralLogicNode): optional, context node from which to build the scene
        """
        scene_name = "/"
        icon = QtGui.QIcon()
        if context:
            scene_name = context.full_name
            icon = QtGui.QIcon(CR.get_icon_path(context.class_name))

        graphics_view = graphic_scene.CustomGraphicsView()
        graphics_scene = graphic_scene.CustomScene(context)
        graphics_view.setScene(graphics_scene)

        graphics_scene.setSceneRect(-20000, -20000, 40000, 40000)
        graphics_scene.setParent(graphics_view)
        GS.signals.main_screen_feedback.connect(graphics_view.show_feedback)

        if context:
            graphics_scene.load_from_file(context.CONTEXT_DEFINITION_FILE, False)

        self.ui.tabWidget.addTab(graphics_view, icon, scene_name)
        self.ui.tabWidget.setCurrentIndex(self.ui.tabWidget.count() - 1)

    def expand_context(self, uuid):
        """
        Expand the contents of a context node into a new tab.

        Args:
            uuid (str): of the node to try to expand
        """
        n = self.get_node_by_uuid(uuid)

        # See if it is a context
        if not n.IS_CONTEXT:
            return

        # See if the context is already expanded
        for i in range(self.ui.tabWidget.count()):
            gw = self.ui.tabWidget.widget(i)
            logic_scene = gw.scene().logic_scene
            if logic_scene.context and logic_scene.context == n:
                self.ui.tabWidget.setCurrentIndex(i)
                return

        # Otherwise, expand it
        self.add_scene(n)
        current_gw = self.ui.tabWidget.widget(self.ui.tabWidget.currentIndex())
        current_scene = current_gw.scene()
        current_scene.fit_in_view()

    def refresh_tab_names(self):
        """
        Refresh the names of all tabs, to make sure they show the names they have to.
        """
        for i in range(self.ui.tabWidget.count()):
            gw = self.ui.tabWidget.widget(i)
            logic_scene = gw.scene().logic_scene
            if logic_scene.context:
                self.ui.tabWidget.setTabText(i, logic_scene.context.full_name)
                return

    # NODE UTILITY ----------------------
    def get_node_by_uuid(self, uuid):
        """
        Given a uuid, find a logic node.

        Args:
            uuid (str): of the node to find

        Returns:
            GeneralLogicNode: logic node associated to the uuid

        """
        for i in range(self.ui.tabWidget.count()):
            current_gw = self.ui.tabWidget.widget(i)
            current_logic_scene = current_gw.scene().logic_scene
            for n in current_logic_scene.all_nodes():
                if n.uuid == uuid:
                    return n

    def add_node_to_current(self, pos, class_name=None):
        """
        Add a new node to the current graphic scene.

        Args:
            x (int): horizontal coord t (in view coords) add the node to
            y (int): vertical coord t (in view coords) add the node to
            class_name (str, optional): class name of the node to instantiate
        """
        current_gw = self.ui.tabWidget.widget(self.ui.tabWidget.currentIndex())
        current_scene = current_gw.scene()
        node_class_name = class_name
        item_type = constants.GRAPHIC_NODE

        if class_name is None:
            selected_item = (
                self.ui.nodes_tree.selectedItems()
                or self.ui.annotations_tree.selectedItems()
            )
            if not selected_item:
                return
            selected = selected_item[0]
            node_class_name = str(selected.data(0, QtCore.Qt.UserRole)).strip()
            item_type = str(selected.data(0, QtCore.Qt.UserRole + 1)).strip()

        if item_type == constants.GRAPHIC_NODE:
            current_scene.add_graphic_node_by_class_name(
                node_class_name,
                current_gw.mapToScene(pos).x() - 50,
                current_gw.mapToScene(pos).y() - 20,
            )

        elif item_type == constants.GRAPHIC_ANNOTATION:
            current_scene.add_annotation_by_type(
                node_class_name,
                current_gw.mapToScene(pos).x() - 50,
                current_gw.mapToScene(pos).y() - 50,
            )

    # ATTRIBUTE EDITOR ----------------------
    def refresh_node_in_attribute_editor_by_uuid(self, uuid):
        """
        Refresh a given node's representation in the attribute editor panel.

        Args:
            uuid (str): of the node to be refreshed
        """
        n = self.get_node_by_uuid(uuid)
        self.attr_editor.refresh_node_panel(n)

    def remove_node_in_attribute_editor_by_uuid(self, uuid):
        """
        Remove a given node's representation from the attribute editor panel.

        Args:
            uuid (str): of the node to be removed
        """
        n = self.get_node_by_uuid(uuid)
        self.attr_editor.remove_node_panel(n)

    def add_node_to_attribute_editor_by_uuid(self, uuid):
        """
        Add a given node's representation to the attribute editor panel.

        Args:
            uuid (str): of the node to be added
        """
        n = self.get_node_by_uuid(uuid)
        self.attr_editor.add_node_panel(n)
        self.attr_editr_dock.show()

    # SAVE AND LOAD ----------------------
    def save_scene(self):
        """
        Launch process of saving changes of current scene to a file, or just save as.
        """
        current_gw = self.ui.tabWidget.widget(self.ui.tabWidget.currentIndex())
        current_scene = current_gw.scene()
        if current_scene.filepath:
            current_scene.save_to_file(current_scene.filepath)
        else:
            current_scene.save_to_file()

    def save_scene_as(self):
        """
        Launch process of saving current scene to a file.
        """
        current_gw = self.ui.tabWidget.widget(self.ui.tabWidget.currentIndex())
        current_scene = current_gw.scene()
        current_scene.save_to_file()

    def load_scene(self, source_file):
        """
        Launch the process of loading a file onto a new tab.

        Args:
            source_file (str): path of the scene to load
        """
        # Filepath
        if not source_file:
            dialog = QtWidgets.QFileDialog()
            result = dialog.getOpenFileName(
                caption="Specify source file", filter="*.yml"
            )
            if not result[0] or not result[1]:
                return
            source_file = result[0]

        # Clear scenes
        self.clear_tabs()

        # Clear attr editor
        self.attr_editor.clear_all()

        # Add scene and load
        self.add_scene()
        current_gw = self.ui.tabWidget.widget(self.ui.tabWidget.currentIndex())
        current_scene = current_gw.scene()
        current_scene.load_from_file(source_file)
        current_scene.fit_in_view()
        self.ui.tabWidget.setTabText(
            self.ui.tabWidget.currentIndex(),
            os.path.basename(source_file),
        )
        self.ui.tabWidget.setTabIcon(
            self.ui.tabWidget.currentIndex(), QtGui.QIcon("icons:scene.png")
        )

    # SCENE EXECUTION ----------------------
    def run_current_scene(self):
        """
        Execute the graphic scene currently selected.
        """
        current_gw = self.ui.tabWidget.widget(self.ui.tabWidget.currentIndex())
        current_scene = current_gw.scene()
        current_scene.run_graphic_scene()
        self.attr_editor.refresh()

    def reset_current_scene(self):
        """
        Reset the graphic scene currently selected.
        """
        current_gw = self.ui.tabWidget.widget(self.ui.tabWidget.currentIndex())
        current_scene = current_gw.scene()
        current_scene.reset_graphic_scene()
        self.attr_editor.refresh()

    def show_scene_results(self):
        """
        Display execution results on the nodes of the graphic scene currently selected.
        """
        current_index = self.ui.tabWidget.currentIndex()
        if current_index < 0:
            return
        current_gw = self.ui.tabWidget.widget(current_index)
        current_scene = current_gw.scene()
        current_scene.show_result_on_nodes()
